from .iconfont.font import IconFont  # noqa: F401
from .iconfont.sources import THIRDPARTY_FONTSPEC  # noqa: F401
from .iconfont.fontspec import ICON_ASSETS_DIR  # noqa: F401
from .iconfont.icon import IconSpec, IconStateSpec, IconLayer  # noqa: F401
from .iconfont import init_iconfonts  # noqa: F401



from os import environ

# these constants are only relevant to testing, do not add to constants
ENV_CI_RUNNING = 'CI'
ENV_PYRANDYOS_UNITTEST_ACTIVE = 'PYRANDYOS_UNITTEST_ACTIVE'


def _is_pyrandyos_unittest():
    return bool(environ.get(ENV_PYRANDYOS_UNITTEST_ACTIVE))


def _is_running_in_ci():
    return bool(environ.get(ENV_CI_RUNNING))



from __future__ import annotations
from typing import TYPE_CHECKING
from collections.abc import Callable

from ...logging import log_func_call
from .themes import qdarkstyle, light, vibedark, HAS_QDARKSTYLE
if TYPE_CHECKING:
    from ..gui_app import QtApp

DEFAULT_THEME_NAME = '(default)'
STATUS_LABEL = 'status_label'  # used in stylesheet for status label


class ThemeMap:
    @log_func_call
    def __init__(self, app: QtApp):
        self.app = app
        self.__custom_themes: dict[str, Callable] = dict()
        self.__current: str = None
        self.init_themes()

    @log_func_call
    def create_theme(self, name: str, callback: Callable = None):
        themes = self.__custom_themes
        if name not in themes:
            themes[name.lower()] = callback if callback else self.default_theme

    @log_func_call
    def init_themes(self):
        self.create_theme('light', light)
        self.create_theme('qdarkstyle', qdarkstyle)
        self.create_theme('vibedark', vibedark)

    @log_func_call
    def apply_theme(self, name: str = None):
        if not name or name == DEFAULT_THEME_NAME:
            self.default_theme(self.app)
        else:
            self.__custom_themes[name.lower()](self.app)

        self.__current = name

    @log_func_call
    def get_current_theme(self):
        return self.__current

    @log_func_call
    def list_themes(self, always_include_qdarkstyle: bool = False,
                    include_default: bool = True):
        tmp = list(self.__custom_themes.keys())
        if not HAS_QDARKSTYLE and (not always_include_qdarkstyle
                                   and 'qdarkstyle' in tmp):
            tmp.remove('qdarkstyle')

        if include_default and DEFAULT_THEME_NAME not in tmp:
            tmp.append(DEFAULT_THEME_NAME)

        return tmp

    @classmethod
    @log_func_call
    def default_theme(cls, app: QtApp):
        app.setPalette(app.style().standardPalette())
        app.setStyleSheet('')



from pathlib import Path

from ..logging import log_func_call, DEBUGLOW2, WARNING
from .net import download_file, get_github_download_url
from .filemeta import filehash


class GitCommitSpec:
    @log_func_call(DEBUGLOW2, trace_only=True)
    def __init__(self, git_repo_base_url: str, git_commit_hash: str,
                 license_relpath: Path | tuple[Path] = None):
        self.git_repo_base_url = git_repo_base_url
        self.git_commit_hash = git_commit_hash
        self.license_relpath = license_relpath


class GitFileSpec:
    @log_func_call(DEBUGLOW2, trace_only=True)
    def __init__(self, git_commit: GitCommitSpec, repo_relpath: Path,
                 md5sum: str = None, local_path: Path = None):
        self.git_commit = git_commit
        self.repo_relpath = repo_relpath
        self.md5sum = md5sum
        self.local_path = local_path
        self.parent: 'GitDependencySpec' = None

    @log_func_call(DEBUGLOW2, trace_only=True)
    def get_local_path(self, download_dir: Path | None = None,
                       override_path: Path | None = None):
        name = self.repo_relpath.name
        p = self.local_path or override_path
        if not p:
            raise ValueError("Base path must be provided if local_path is "
                             "not set")
        if p.is_dir():
            p /= name

        if p.exists():
            return p

        if download_dir is None:
            from ..app import PyRandyOSApp
            download_dir = PyRandyOSApp.mkdir_temp()

        download_dir.mkdir(parents=True, exist_ok=True)
        return download_dir/name

    @log_func_call(WARNING)
    def download(self, dest: Path, use_tqdm: bool = True,
                 show_full_path: bool = False):
        git = self.git_commit
        repo = git.git_repo_base_url
        commit = git.git_commit_hash
        relpath = self.repo_relpath
        return download_file(get_github_download_url(repo, commit, relpath),
                             dest, use_tqdm, show_full_path)

    @log_func_call(DEBUGLOW2, trace_only=True)
    def get_or_download(self, download_dir: Path = None, use_tqdm: bool = True,
                        show_full_path: bool = False):
        p = self.get_local_path(download_dir)
        if not p.exists():
            p = self.download(p, use_tqdm, show_full_path)
            if not p.exists():
                raise FileNotFoundError(f"Failed to find or download {p}")

        if self.md5sum and self.md5sum != filehash(p, algorithm='md5'):
            raise ValueError(f"MD5 checksum does not match for {p}")
        return p

    @log_func_call
    def get_or_download_licenses(self, download_dir: Path = None,
                                 use_tqdm: bool = True,
                                 show_full_path: bool = True):
        # get the directory of the corresponding main file
        # also serves as the download dir for get_or_download later here
        p = self.get_local_path(download_dir).parent
        gitcommit = self.git_commit
        license_paths = gitcommit.license_relpath
        if not license_paths:
            return
        if not isinstance(license_paths, tuple):
            license_paths = (license_paths,)

        return tuple(GitFileSpec(gitcommit, lic,
                                 local_path=p).get_or_download(p, use_tqdm,
                                                               show_full_path)
                     for lic in license_paths)


class GitDependencySpec:
    pass



# The MIT License
#
# Copyright (c) 2015 The Spyder development team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from ...qt import QThread, QFontDatabase, QFont, QRawFont
from ....logging import log_func_call, DEBUGLOW2
from .sources import THIRDPARTY_FONTSPEC

_DEFHINT = QFont.PreferDefaultHinting


class IconFontNotInitializedError(RuntimeError):
    pass


class IconFontCacheEntry:
    @log_func_call(DEBUGLOW2, trace_only=True)
    def __init__(self):
        self.ttf_data: bytes = None
        self.id_: int | None = None
        self.font_name: str = None
        self.rawfont: dict[str, QRawFont] = dict()


class IconFontMeta(type):
    @log_func_call(DEBUGLOW2, trace_only=True)
    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)
        cls._cache = IconFontCacheEntry()


class IconFont(metaclass=IconFontMeta):
    _SPECNAME: str

    @classmethod
    @log_func_call(DEBUGLOW2, trace_only=True)
    def get_spec(cls):
        return THIRDPARTY_FONTSPEC[cls._SPECNAME]

    @log_func_call
    def __init__(self):
        self.ensure_font_loaded()

    @classmethod
    @log_func_call(DEBUGLOW2, trace_only=True)
    def get_font(cls, size: float | int):
        "Return a QFont corresponding to the given prefix and size."
        cls.ensure_font_loaded()
        spec = cls.get_spec()
        cache = cls._cache
        font = QFont()
        font.setFamily(cache.font_name)
        font.setPixelSize(round(size))
        if spec.solid:  # solid style
            font.setStyleName("Solid")
        return font

    @classmethod
    @log_func_call(DEBUGLOW2, trace_only=True)
    def get_rawfont(cls, size: float | int,
                    hintingpref: QFont.HintingPreference = _DEFHINT):
        cls.ensure_font_loaded()
        clscache = cls._cache
        cache = clscache.rawfont
        thread = QThread.currentThread()
        tid = str(thread)
        tc = cache.get(tid)
        if not tc:
            tc = dict()
            cache[tid] = tc
            thread.finished.connect(lambda: cache.pop(tid, None))

        key = (size, hintingpref)
        rawfont = tc.get(key)
        if not rawfont:
            rawfont = QRawFont(clscache.ttf_data, size, hintingpref)
            tc[key] = rawfont

        return rawfont

    @classmethod
    @log_func_call(DEBUGLOW2, trace_only=True)
    def ensure_font_loaded(cls):
        cache = cls._cache
        if cache.id_ is None:
            spec = cls.get_spec()
            ttf_path = spec.ttf_filespec.get_local_path()
            fontdata = ttf_path.read_bytes()
            id_ = QFontDatabase.addApplicationFontFromData(fontdata)
            loadedFontFamilies = QFontDatabase.applicationFontFamilies(id_)
            if loadedFontFamilies:
                cache.ttf_data = fontdata
                cache.id_ = id_
                cache.font_name = loadedFontFamilies[0]
            else:
                raise RuntimeError(
                    f"Font '{ttf_path}' appears to be empty. "
                    "If you are on Windows 10, please read "
                    "https://support.microsoft.com/en-us/kb/3053676 "
                    "to know how to prevent Windows from blocking "
                    "the fonts that come with the package."
                )

    @classmethod
    @log_func_call(DEBUGLOW2, trace_only=True)
    def get_codepoint_by_name(cls, icon_name: str):
        charmap = cls.get_spec().charmap
        if charmap:
            return charmap.get(icon_name, None)
        raise IconFontNotInitializedError("Qt application must be initialized "
                                          "before using icon font methods "
                                          "that require the fonts to "
                                          "be loaded.")

    @classmethod
    @log_func_call(DEBUGLOW2, trace_only=True)
    def get_glyph(cls, name_or_codepoint: str | int):
        i = name_or_codepoint
        if isinstance(name_or_codepoint, str):
            i = cls.get_codepoint_by_name(name_or_codepoint)
            if i is None:
                raise ValueError(f"Glyph '{name_or_codepoint}' not found "
                                 "in charmap.")
        return chr(i)



# This file is auto-generated by pyrandyos/gui/icons/update_from_spec.py  # noqa: E501
# Do not edit.
add = 60000
plus = 60000
gist_new = 60000
repo_create = 60000
lightbulb = 60001
light_bulb = 60001
repo = 60002
repo_delete = 60002
gist_fork = 60003
repo_forked = 60003
git_pull_request = 60004
git_pull_request_abandoned = 60004
record_keys = 60005
keyboard = 60005
tag = 60006
git_pull_request_label = 60006
tag_add = 60006
tag_remove = 60006
person = 60007
person_follow = 60007
person_outline = 60007
person_filled = 60007
git_branch = 60008
git_branch_create = 60008
git_branch_delete = 60008
source_control = 60008
mirror = 60009
mirror_public = 60009
star = 60010
star_add = 60010
star_delete = 60010
star_empty = 60010
comment = 60011
comment_add = 60011
alert = 60012
warning = 60012
search = 60013
search_save = 60013
log_out = 60014
sign_out = 60014
log_in = 60015
sign_in = 60015
eye = 60016
eye_unwatch = 60016
eye_watch = 60016
circle_filled = 60017
primitive_dot = 60017
close_dirty = 60017
debug_breakpoint = 60017
debug_breakpoint_disabled = 60017
debug_hint = 60017
terminal_decoration_success = 60017
primitive_square = 60018
edit = 60019
pencil = 60019
info = 60020
issue_opened = 60020
gist_private = 60021
git_fork_private = 60021
lock = 60021
mirror_private = 60021
close = 60022
remove_close = 60022
x = 60022
repo_sync = 60023
sync = 60023
clone = 60024
desktop_download = 60024
beaker = 60025
microscope = 60025
vm = 60026
device_desktop = 60026
file = 60027
file_text = 60027
more = 60028
ellipsis = 60028
kebab_horizontal = 60028
mail_reply = 60029
reply = 60029
organization = 60030
organization_filled = 60030
organization_outline = 60030
new_file = 60031
file_add = 60031
new_folder = 60032
file_directory_create = 60032
trash = 60033
trashcan = 60033
history = 60034
clock = 60034
folder = 60035
file_directory = 60035
symbol_folder = 60035
logo_github = 60036
mark_github = 60036
github = 60036
terminal = 60037
console = 60037
repl = 60037
zap = 60038
symbol_event = 60038
error = 60039
stop = 60039
variable = 60040
symbol_variable = 60040
array = 60042
symbol_array = 60042
symbol_module = 60043
symbol_package = 60043
symbol_namespace = 60043
symbol_object = 60043
symbol_method = 60044
symbol_function = 60044
symbol_constructor = 60044
symbol_boolean = 60047
symbol_null = 60047
symbol_numeric = 60048
symbol_number = 60048
symbol_structure = 60049
symbol_struct = 60049
symbol_parameter = 60050
symbol_type_parameter = 60050
symbol_key = 60051
symbol_text = 60051
symbol_reference = 60052
go_to_file = 60052
symbol_enum = 60053
symbol_value = 60053
symbol_ruler = 60054
symbol_unit = 60054
activate_breakpoints = 60055
archive = 60056
arrow_both = 60057
arrow_down = 60058
arrow_left = 60059
arrow_right = 60060
arrow_small_down = 60061
arrow_small_left = 60062
arrow_small_right = 60063
arrow_small_up = 60064
arrow_up = 60065
bell = 60066
bold = 60067
book = 60068
bookmark = 60069
debug_breakpoint_conditional_unverified = 60070
debug_breakpoint_conditional = 60071
debug_breakpoint_conditional_disabled = 60071
debug_breakpoint_data_unverified = 60072
debug_breakpoint_data = 60073
debug_breakpoint_data_disabled = 60073
debug_breakpoint_log_unverified = 60074
debug_breakpoint_log = 60075
debug_breakpoint_log_disabled = 60075
briefcase = 60076
broadcast = 60077
browser = 60078
bug = 60079
calendar = 60080
case_sensitive = 60081
check = 60082
checklist = 60083
chevron_down = 60084
chevron_left = 60085
chevron_right = 60086
chevron_up = 60087
chrome_close = 60088
chrome_maximize = 60089
chrome_minimize = 60090
chrome_restore = 60091
circle_outline = 60092
circle = 60092
debug_breakpoint_unverified = 60092
terminal_decoration_incomplete = 60092
circle_slash = 60093
circuit_board = 60094
clear_all = 60095
clippy = 60096
close_all = 60097
cloud_download = 60098
cloud_upload = 60099
code = 60100
collapse_all = 60101
color_mode = 60102
comment_discussion = 60103
credit_card = 60105
dash = 60108
dashboard = 60109
database = 60110
debug_continue = 60111
debug_disconnect = 60112
debug_pause = 60113
debug_restart = 60114
debug_start = 60115
debug_step_into = 60116
debug_step_out = 60117
debug_step_over = 60118
debug_stop = 60119
debug = 60120
device_camera_video = 60121
device_camera = 60122
device_mobile = 60123
diff_added = 60124
diff_ignored = 60125
diff_modified = 60126
diff_removed = 60127
diff_renamed = 60128
diff = 60129
diff_sidebyside = 60129
discard = 60130
editor_layout = 60131
empty_window = 60132
exclude = 60133
extensions = 60134
eye_closed = 60135
file_binary = 60136
file_code = 60137
file_media = 60138
file_pdf = 60139
file_submodule = 60140
file_symlink_directory = 60141
file_symlink_file = 60142
file_zip = 60143
files = 60144
filter = 60145
flame = 60146
fold_down = 60147
fold_up = 60148
fold = 60149
folder_active = 60150
folder_opened = 60151
gear = 60152
gift = 60153
gist_secret = 60154
gist = 60155
git_commit = 60156
git_compare = 60157
compare_changes = 60157
git_merge = 60158
github_action = 60159
github_alt = 60160
globe = 60161
grabber = 60162
graph = 60163
gripper = 60164
heart = 60165
home = 60166
horizontal_rule = 60167
hubot = 60168
inbox = 60169
issue_reopened = 60171
issues = 60172
italic = 60173
jersey = 60174
json = 60175
kebab_vertical = 60176
key = 60177
law = 60178
lightbulb_autofix = 60179
link_external = 60180
link = 60181
list_ordered = 60182
list_unordered = 60183
live_share = 60184
loading = 60185
location = 60186
mail_read = 60187
mail = 60188
markdown = 60189
megaphone = 60190
mention = 60191
milestone = 60192
git_pull_request_milestone = 60192
mortar_board = 60193
move = 60194
multiple_windows = 60195
mute = 60196
no_newline = 60197
note = 60198
octoface = 60199
open_preview = 60200
package = 60201
paintcan = 60202
pin = 60203
play = 60204
run = 60204
plug = 60205
preserve_case = 60206
preview = 60207
project = 60208
pulse = 60209
question = 60210
quote = 60211
radio_tower = 60212
reactions = 60213
references = 60214
refresh = 60215
regex = 60216
remote_explorer = 60217
remote = 60218
remove = 60219
replace_all = 60220
replace = 60221
repo_clone = 60222
repo_force_push = 60223
repo_pull = 60224
repo_push = 60225
report = 60226
request_changes = 60227
rocket = 60228
root_folder_opened = 60229
root_folder = 60230
rss = 60231
ruby = 60232
save_all = 60233
save_as = 60234
save = 60235
screen_full = 60236
screen_normal = 60237
search_stop = 60238
server = 60240
settings_gear = 60241
settings = 60242
shield = 60243
smiley = 60244
sort_precedence = 60245
split_horizontal = 60246
split_vertical = 60247
squirrel = 60248
star_full = 60249
star_half = 60250
symbol_class = 60251
symbol_color = 60252
symbol_constant = 60253
symbol_enum_member = 60254
symbol_field = 60255
symbol_file = 60256
symbol_interface = 60257
symbol_keyword = 60258
symbol_misc = 60259
symbol_operator = 60260
symbol_property = 60261
wrench = 60261
wrench_subaction = 60261
symbol_snippet = 60262
tasklist = 60263
telescope = 60264
text_size = 60265
three_bars = 60266
thumbsdown = 60267
thumbsup = 60268
tools = 60269
triangle_down = 60270
triangle_left = 60271
triangle_right = 60272
triangle_up = 60273
twitter = 60274
unfold = 60275
unlock = 60276
unmute = 60277
unverified = 60278
verified = 60279
versions = 60280
vm_active = 60281
vm_outline = 60282
vm_running = 60283
watch = 60284
whitespace = 60285
whole_word = 60286
window = 60287
word_wrap = 60288
zoom_in = 60289
zoom_out = 60290
list_filter = 60291
list_flat = 60292
list_selection = 60293
selection = 60293
list_tree = 60294
debug_breakpoint_function_unverified = 60295
debug_breakpoint_function = 60296
debug_breakpoint_function_disabled = 60296
debug_stackframe_active = 60297
circle_small_filled = 60298
debug_stackframe_dot = 60298
terminal_decoration_mark = 60298
debug_stackframe = 60299
debug_stackframe_focused = 60299
debug_breakpoint_unsupported = 60300
symbol_string = 60301
debug_reverse_continue = 60302
debug_step_back = 60303
debug_restart_frame = 60304
debug_alt = 60305
call_incoming = 60306
call_outgoing = 60307
menu = 60308
expand_all = 60309
feedback = 60310
git_pull_request_reviewer = 60310
group_by_ref_type = 60311
ungroup_by_ref_type = 60312
account = 60313
git_pull_request_assignee = 60313
bell_dot = 60314
debug_console = 60315
library = 60316
output = 60317
run_all = 60318
sync_ignored = 60319
pinned = 60320
github_inverted = 60321
server_process = 60322
server_environment = 60323
pass_ = 60324
issue_closed = 60324
stop_circle = 60325
play_circle = 60326
record = 60327
debug_alt_small = 60328
vm_connect = 60329
cloud = 60330
merge = 60331
export = 60332
graph_left = 60333
magnet = 60334
notebook = 60335
redo = 60336
check_all = 60337
pinned_dirty = 60338
pass_filled = 60339
circle_large_filled = 60340
circle_large = 60341
circle_large_outline = 60341
combine = 60342
gather = 60342
table = 60343
variable_group = 60344
type_hierarchy = 60345
type_hierarchy_sub = 60346
type_hierarchy_super = 60347
git_pull_request_create = 60348
run_above = 60349
run_below = 60350
notebook_template = 60351
debug_rerun = 60352
workspace_trusted = 60353
workspace_untrusted = 60354
workspace_unknown = 60355
terminal_cmd = 60356
terminal_debian = 60357
terminal_linux = 60358
terminal_powershell = 60359
terminal_tmux = 60360
terminal_ubuntu = 60361
terminal_bash = 60362
arrow_swap = 60363
copy = 60364
person_add = 60365
filter_filled = 60366
wand = 60367
debug_line_by_line = 60368
inspect = 60369
layers = 60370
layers_dot = 60371
layers_active = 60372
compass = 60373
compass_dot = 60374
compass_active = 60375
azure = 60376
issue_draft = 60377
git_pull_request_closed = 60378
git_pull_request_draft = 60379
debug_all = 60380
debug_coverage = 60381
run_errors = 60382
folder_library = 60383
debug_continue_small = 60384
beaker_stop = 60385
graph_line = 60386
graph_scatter = 60387
pie_chart = 60388
bracket = 60175
bracket_dot = 60389
bracket_error = 60390
lock_small = 60391
azure_devops = 60392
verified_filled = 60393
newline = 60394
layout = 60395
layout_activitybar_left = 60396
layout_activitybar_right = 60397
layout_panel_left = 60398
layout_panel_center = 60399
layout_panel_justify = 60400
layout_panel_right = 60401
layout_panel = 60402
layout_sidebar_left = 60403
layout_sidebar_right = 60404
layout_statusbar = 60405
layout_menubar = 60406
layout_centered = 60407
target = 60408
indent = 60409
record_small = 60410
error_small = 60411
terminal_decoration_error = 60411
arrow_circle_down = 60412
arrow_circle_left = 60413
arrow_circle_right = 60414
arrow_circle_up = 60415
layout_sidebar_right_off = 60416
layout_panel_off = 60417
layout_sidebar_left_off = 60418
blank = 60419
heart_filled = 60420
map = 60421
map_horizontal = 60421
fold_horizontal = 60421
map_filled = 60422
map_horizontal_filled = 60422
fold_horizontal_filled = 60422
circle_small = 60423
bell_slash = 60424
bell_slash_dot = 60425
comment_unresolved = 60426
git_pull_request_go_to_changes = 60427
git_pull_request_new_changes = 60428
search_fuzzy = 60429
comment_draft = 60430
send = 60431
sparkle = 60432
insert = 60433
mic = 60434
thumbsdown_filled = 60435
thumbsup_filled = 60436
coffee = 60437
snake = 60438
game = 60439
vr = 60440
chip = 60441
piano = 60442
music = 60443
mic_filled = 60444
repo_fetch = 60445
copilot = 60446
lightbulb_sparkle = 60447
robot = 60448
sparkle_filled = 60449
diff_single = 60450
diff_multiple = 60451
surround_with = 60452
share = 60453
git_stash = 60454
git_stash_apply = 60455
git_stash_pop = 60456
vscode = 60457
vscode_insiders = 60458
code_oss = 60459
run_coverage = 60460
run_all_coverage = 60461
coverage = 60462
github_project = 60463
map_vertical = 60464
fold_vertical = 60464
map_vertical_filled = 60465
fold_vertical_filled = 60465
go_to_search = 60466
percentage = 60467
sort_percentage = 60467
attach = 60468
go_to_editing_session = 60469
edit_session = 60470
code_review = 60471
copilot_warning = 60472
python = 60473
copilot_large = 60474
copilot_warning_large = 60475
keyboard_tab = 60476
copilot_blocked = 60477
copilot_not_connected = 60478
flag = 60479
lightbulb_empty = 60480
symbol_method_arrow = 60481
copilot_unavailable = 60482
repo_pinned = 60483
keyboard_tab_above = 60484
keyboard_tab_below = 60485
git_pull_request_done = 60486
mcp = 60487
extensions_large = 60488
layout_panel_dock = 60489
layout_sidebar_left_dock = 60490
layout_sidebar_right_dock = 60491
copilot_in_progress = 60492
copilot_error = 60493
copilot_success = 60494
chat_sparkle = 60495
search_sparkle = 60496
edit_sparkle = 60497
copilot_snooze = 60498



from .constants.cli import CSI
from ..logging import log_func_call


class ScreenControl:
    @staticmethod
    @log_func_call
    def put_cursor_at_line_col(line: int, col: int):
        # return ff'{CSI}{line};{col}f'
        return f'{CSI}{line};{col}H'

    @staticmethod
    @log_func_call
    def move_cursor_up(n: int):
        return f'{CSI}{n}A'

    @staticmethod
    @log_func_call
    def move_cursor_down(n: int):
        return f'{CSI}{n}B'

    @staticmethod
    @log_func_call
    def move_cursor_fwd(n: int):
        return f'{CSI}{n}C'

    @staticmethod
    @log_func_call
    def move_cursor_bwd(n: int):
        return f'{CSI}{n}D'


@log_func_call
def render_console_string(s: str):
    """
    parses a string with ANSI control characters and generates the equivalent
    after all characters are played
    """
    lines = ['']
    col = 0
    row = 0
    esc = None
    mcol = None
    mrow = None
    for c in s:
        if esc is not None:
            if (esc + c) == CSI:
                esc = ''
            elif c in 'ABCDHfmKJsu':
                if c in 'ABCD':
                    esc = int(esc) if esc else 1
                    if c == 'A':
                        row -= esc

                    elif c == 'B':
                        row += esc

                    elif c == 'C':
                        col += esc

                    elif c == 'D':
                        col -= esc

                elif c == 'K':
                    esc = int(esc) if esc else 0
                    # if esc == 0:
                    #     pass  # clear to EOL
                    #
                    # elif esc == 1:
                    #     pass  # clear to BOL
                    #
                    # elif esc == 2:
                    #     pass  # clear line
                    raise NotImplementedError

                elif c == 'J':
                    esc = int(esc) if esc else 0
                    if esc == 2:
                        lines = ['']
                        col = 0
                        row = 0

                    else:
                        raise NotImplementedError

                elif c == 's':
                    mcol = col
                    mrow = row

                elif c == 'u':
                    col = mcol
                    row = mrow

                elif c in 'Hf':
                    row, col = esc.split(';')
                    row = int(row) - 1
                    col = int(col) - 1

                esc = None

            else:
                esc += c

        elif c == '\033':
            esc = c

        elif c == '\r':
            col = 0

        elif c == '\n':
            row += 1
            col = 0
            if row == len(lines):
                lines.append('')

        else:
            linediff = row - len(lines) + 1
            if linediff > 0:
                lines += ['']*linediff

            old = lines[row]
            n = len(old)
            if col > n:
                old = ' '*col
            lines[row] = old[:col] + c + old[col + 1:]
            col += 1

    return '\n'.join(lines)



from pathlib import Path

from ....app import PyRandyOSApp
from ....config.keys import BASE_LOG_PATH_KEY
from ....logging import log_func_call

from .pres import LogDialog


class LogDialogMainWindowMixin:
    @log_func_call
    def show_log_dialog(self, log_path: Path = None):
        log_path = log_path or self.get_log_path()
        if log_path:
            # Close existing dialog if open
            dialog: LogDialog = self._log_dialog
            if dialog:
                try:
                    dialog.gui_view.qtobj.close()
                except RuntimeError:
                    # Dialog was already closed/destroyed
                    pass

            # Create and show new dialog
            dialog = LogDialog(self, log_path)
            self._log_dialog = dialog
            dialog.show()

    @log_func_call
    def get_log_path(self):
        logpath = PyRandyOSApp.get(BASE_LOG_PATH_KEY, None)
        return Path(logpath) if logpath else None



import sys
from types import TracebackType

from ..logging import log_func_call, log_exc, log_debug, log_info
from .._testing.debug import is_debug_enabled
from .log import setup_memory_logging
from .stack import exc_info


class MainContext:
    """
    Context manager for the main application context.
    It sets up logging and handles exceptions.
    """

    @log_func_call
    def __init__(self, appname: str):
        self.appname = appname

    @log_func_call
    def __enter__(self):
        try:
            setup_memory_logging()
            log_debug(f"starting {self.appname} main")
        except BaseException as e:
            self.__exit__(*exc_info(exc=e))

        return self

    @log_func_call
    def __exit__(self, exc_type: type = None, exc: BaseException = None,
                 traceback: TracebackType = None):
        exit_code = exc.code if isinstance(exc, SystemExit) else 0
        status_ok = True
        try:
            if exc:
                if not isinstance(exc, (SystemExit, KeyboardInterrupt)):
                    status_ok = False
                    exit_code = 1
                    log_exc(exc_type, exc, traceback)

                    # if there is a splash screen, kill it
                    try:
                        from ..gui.gui_app import get_gui_app
                        gui_app = get_gui_app()
                        if gui_app:
                            gui_app.splash.gui_view.qtobj.hide()

                    except BaseException:
                        pass

                if is_debug_enabled():
                    # this triggers the debugger to break on the exception
                    raise exc

        finally:
            appname = self.appname
            if exit_code:
                log_info(f'{appname} exited with code {exit_code}')

            if status_ok:
                log_info(f'{appname} exiting gracefully')
            else:
                pass

            if not is_debug_enabled():
                # if running in the debugger, we don't want to catch the
                # SystemExit so we can see the traceback in the console more
                # clearly. However, this means that the debugger will always
                # exit with code 1 if an unhandled exception, even if we
                # specify a different exit code above since we only call
                # sys.exit(exit_code) here in the case where the debugger is
                # not enabled.  Ideally, we would set the exit code in the
                # debugger as well, but that is not possible with the current
                # setup without modifying the launch configuration.  I prefer
                # to keep this simple and rely on the log messages below to
                # indicate the "true" exit code when running in the debugger.
                sys.exit(exit_code)



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    def get_ipython():
        pass

from ..logging import log_func_call, DEBUGLOW


@log_func_call
def is_notebook() -> bool:  # pragma: no cover
    return get_interpreter() == 'notebook'


@log_func_call(DEBUGLOW)
def get_ipython_if_running():
    try:
        return get_ipython()
    except NameError:
        return None


@log_func_call
def get_interpreter():
    # adapted from: https://stackoverflow.com/a/39662359/13230486
    ipy = get_ipython_if_running()
    if ipy:
        cls = ipy.__class__
        shell = cls.__name__
        if shell == 'ZMQInteractiveShell':
            return 'notebook'   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return 'ipython'  # Terminal running IPython
        elif cls.__module__ == "google.colab._shell":
            # https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook#comment93642570_39662359  # noqa: E501
            return 'google'  # Google Colab...but maybe should be true?
        return 'unknown'  # Other type (?)
    return 'python'  # Probably standard Python interpreter



import sys
from typing import TYPE_CHECKING
from collections.abc import Iterable, Iterator, Generator
from pathlib import Path
from contextlib import contextmanager
from unittest.mock import MagicMock

from tqdm.auto import tqdm
from tqdm.utils import _screen_shape_wrapper

from ..logging import log_func_call
from .string import iterable_max_chars
if TYPE_CHECKING:
    from .filemeta import FileSet

TQDM_RBAR_STATWIDTH = 50


@log_func_call
def tqdm_fixed_label_barfmt(x: Iterable | int = None,
                            ncols: int | None = None):
    maxlen = x if isinstance(x, int) else iterable_max_chars(x)
    statwidth = TQDM_RBAR_STATWIDTH + maxlen
    if ncols and statwidth > 0.75*ncols:
        statwidth = int(0.75*ncols)
    return f'{{l_bar}}{{bar}}{{r_bar:{statwidth}}}'


@log_func_call
def tqdm_fixed_label_width(x: Iterable, maxlen: int = None, **kwargs):
    ncols = get_tqdm_ncols(**kwargs)
    return tqdm(x, bar_format=tqdm_fixed_label_barfmt(maxlen or x, ncols),
                **kwargs)


class FileSetTqdm(tqdm):
    """
    A tqdm subclass that can handle FileSet objects, printing the file paths
    in the tqdm description after each iteration.
    """
    @log_func_call
    def __init__(self, fset: 'FileSet', maxlen: int = None, **kwargs):
        # in case fset is a generator, let's preconvert it to a sorted tuple
        fset = sorted(tuple(fset))
        if not maxlen:
            from .filemeta import fileset_max_chars
            maxlen = fileset_max_chars(fset)

        ncols = get_tqdm_ncols(**kwargs)
        fmt = tqdm_fixed_label_barfmt(maxlen, ncols)
        super().__init__(fset, bar_format=fmt, **kwargs)

    @log_func_call
    def update_file(self, p: Path):
        self.set_postfix(file=p.as_posix())
        self.refresh()

    async def __anext__(self):
        res: Path = super().__anext__()
        self.update_file(res)
        return res

    def __iter__(self):
        """Backward-compatibility to use: for x in tqdm(iterable)"""

        the_iter: Iterator[Path] = super().__iter__()
        for obj in the_iter:
            obj: Path
            self.update_file(obj)
            yield obj


@log_func_call
def get_tqdm_ncols(file=None, ncols: int | None = None,
                   dynamic_ncols: bool = False, **kwargs):
    if ncols is not None:
        return ncols

    if file is None:
        file = sys.stderr

    if (file in (sys.stderr, sys.stdout)) or dynamic_ncols:
        if dynamic_ncols:
            dynamic_ncols = _screen_shape_wrapper()
            if dynamic_ncols:
                ncols, _ = dynamic_ncols(file)
        else:
            _dynamic_ncols = _screen_shape_wrapper()
            if _dynamic_ncols:
                ncols, _ = _dynamic_ncols(file)

    return ncols


@contextmanager
@log_func_call
def optional_tqdm(use_tqdm: bool = True, tqdmclass: type = tqdm,
                  *args, **kwargs) -> Generator[tqdm, None, None]:
    """
    Context manager that yields a tqdm instance if use_tqdm is True,
    otherwise yields a MagicMock that no-ops all tqdm methods.
    """
    if use_tqdm:
        bar: tqdm = tqdmclass(*args, **kwargs)
    else:
        bar = MagicMock()
        bar.__enter__.return_value = bar
        bar.__exit__.return_value = None
    try:
        yield bar
    finally:
        bar.close()



# from ..logging import log_func_call
# see https://stackoverflow.com/a/33206814/13230486
CSI = '\033['  # https://en.wikipedia.org/wiki/ANSI_escape_code#CSIsection


class ConsoleText:
    Reset = f'{CSI}0m'
    ResetText = f'{CSI}39m'
    ResetBg = f'{CSI}49m'

    Bold = f'{CSI}1m'
    NoBold = f'{CSI}21m'
    Underline = f'{CSI}4m'
    NoUnderline = f'{CSI}24m'
    Strikethrough = f'{CSI}4m'
    NoStrikethrough = f'{CSI}24m'

    GrayText = f'{CSI}90m'
    RedText = f'{CSI}91m'
    GreenText = f'{CSI}92m'
    YellowText = f'{CSI}93m'
    BlueText = f'{CSI}94m'
    MagentaText = f'{CSI}95m'
    CyanText = f'{CSI}96m'
    WhiteText = f'{CSI}97m'

    BlackText = f'{CSI}30m'
    DarkRedText = f'{CSI}31m'
    DarkGreenText = f'{CSI}32m'
    DarkYellowText = f'{CSI}33m'
    DarkBlueText = f'{CSI}34m'
    DarkMagentaText = f'{CSI}35m'
    DarkCyanText = f'{CSI}36m'
    DarkWhiteText = f'{CSI}37m'

    GrayBg = f'{CSI}100m'
    RedBg = f'{CSI}101m'
    GreenBg = f'{CSI}102m'
    YellowBg = f'{CSI}103m'
    BlueBg = f'{CSI}104m'
    MagentaBg = f'{CSI}105m'
    CyanBg = f'{CSI}106m'
    WhiteBg = f'{CSI}107m'

    BlackBg = f'{CSI}40m'
    DarkRedBg = f'{CSI}41m'
    DarkGreenBg = f'{CSI}42m'
    DarkYellowBg = f'{CSI}43m'
    DarkBlueBg = f'{CSI}44m'
    DarkMagentaBg = f'{CSI}45m'
    DarkCyanBg = f'{CSI}46m'
    DarkWhiteBg = f'{CSI}47m'


class ScreenControlCmds:
    ClearLine = f'{CSI}2K'
    ClearScreen = f'{CSI}2J'
    EraseToEOL = f'{CSI}K'
    SaveCursor = f'{CSI}s'
    RestoreCursor = f'{CSI}u'



from pathlib import Path
from json import loads as jloads
from importlib import import_module

from ....logging import log_func_call, DEBUGLOW2
from ....utils.git import GitCommitSpec, GitFileSpec, GitDependencySpec
from .. import thirdparty

HERE = Path(__file__).parent
ICON_ASSETS_DIR = HERE.parent/'assets'
# THIRDPARTY_DIR = HERE.parent/'thirdparty'
THIRDPARTY_DIR = Path(thirdparty.__file__).parent
CharMap = dict[str, int]


class IconFontGitCommit(GitCommitSpec):
    pass


class IconFontGitFile(GitFileSpec):
    @log_func_call(DEBUGLOW2, trace_only=True)
    def __init__(self, filetype: str, git_commit: IconFontGitCommit,
                 repo_relpath: Path, md5sum: str = None):
        super().__init__(git_commit, repo_relpath, md5sum)
        self.parent: 'IconFontSpec'
        self.filetype = filetype

    @log_func_call(DEBUGLOW2, trace_only=True)
    def get_local_path(self, download_dir: Path | None = None):
        classname = self.parent.classname
        filetype = self.filetype
        p = ICON_ASSETS_DIR/classname/filetype

        if download_dir is None:
            from ....app import PyRandyOSApp
            download_dir = PyRandyOSApp.mkdir_temp()

        return super().get_local_path(download_dir/classname/filetype, p)


class IconTtfFileSpec(IconFontGitFile):
    @log_func_call(DEBUGLOW2, trace_only=True)
    def __init__(self, git_commit: IconFontGitCommit, repo_relpath: Path,
                 md5sum: str = None):
        super().__init__('ttf', git_commit, repo_relpath, md5sum)
        self.parent: 'IconFontSpec'


class IconCharMapFileSpec(IconFontGitFile):
    @log_func_call(DEBUGLOW2, trace_only=True)
    def __init__(self, git_commit: IconFontGitCommit, repo_relpath: Path,
                 codepoint_base: int = 16, md5sum: str = None):
        super().__init__('charmap', git_commit, repo_relpath, md5sum)
        self.codepoint_base = codepoint_base

    @log_func_call
    def load_charmap(self) -> CharMap:
        jsonfile = self.get_local_path()
        charmap: dict[str, str | int] = jloads(jsonfile.read_text())
        cpbase = self.codepoint_base
        return {k: int(v, cpbase) if not isinstance(v, int) else v
                for k, v in charmap.items()}


class IconFontSpec(GitDependencySpec):
    @log_func_call(DEBUGLOW2, trace_only=True)
    def __init__(self,  # target_relative_module_name: str,
                 ttf_filespec: IconTtfFileSpec,
                 charmap_filespec: IconCharMapFileSpec,
                 solid: bool = False):
        self.target_relative_class_qualname = None

        self.ttf_filespec = ttf_filespec
        ttf_filespec.parent = self

        self.charmap_filespec = charmap_filespec
        charmap_filespec.parent = self

        self.solid = solid
        self.charmap: CharMap = None
        self.classname: str = None
        self.shortname: str = None
        self.relative_module_qualname: str = None
        self._initialized = False

    @log_func_call(DEBUGLOW2, trace_only=True)
    def ensure_local_files(self, download_dir: Path = None):
        ttfspec = self.ttf_filespec
        ttffile = ttfspec.get_or_download(download_dir)
        ttflicenses = ttfspec.get_or_download_licenses(download_dir)

        charmapspec = self.charmap_filespec
        charmapfile = charmapspec.get_or_download(download_dir)
        charmaplicenses = charmapspec.get_or_download_licenses(download_dir)

        return ttffile, charmapfile, ttflicenses, charmaplicenses

    @log_func_call(DEBUGLOW2, trace_only=True)
    def initialize(self, target_relative_class_qualname: str,
                   download_dir: Path = None):
        if not self._initialized:
            tmpmodname = target_relative_class_qualname
            self.target_relative_class_qualname = tmpmodname
            self.relative_module_qualname = tmpmodname.lower()
            tmpmodname_parts = tmpmodname.split('.')
            classname = '_'.join(tmpmodname_parts)
            self.classname = classname
            # self.ttf_filespec.classname = classname
            shortnameparts = tmpmodname_parts.copy()
            if len(shortnameparts) > 1:
                shortnameparts[-1] = shortnameparts[-1][0]
            self.shortname = '_'.join(shortnameparts).lower()

            self.ensure_local_files(download_dir)
            self.charmap = self.charmap_filespec.load_charmap()
            self._initialized = True

    @log_func_call(DEBUGLOW2, trace_only=True)
    def relative_class_qualname(self):
        modname = self.relative_module_qualname
        classname = self.classname
        return f"{modname}.{classname}"

    @log_func_call(DEBUGLOW2, trace_only=True)
    def module_qualname(self):
        thirdpartymod = thirdparty.__name__
        relqualname = self.relative_module_qualname
        return f'{thirdpartymod}.{relqualname}'

    @log_func_call(DEBUGLOW2, trace_only=True)
    def import_font(self):
        return import_module(self.module_qualname())

    @log_func_call(DEBUGLOW2, trace_only=True)
    def get_font_class(self):
        mod = self.import_font()
        classname = self.classname
        return getattr(mod, classname)



from pathlib import Path
from ..logging import log_func_call


@log_func_call
def read_text_utf8(path: str | Path) -> str:
    return Path(path).read_text(encoding='utf-8')


@log_func_call
def write_text_utf8(path: str | Path, content: str) -> None:
    Path(path).write_text(content, encoding='utf-8')



import sys


args = sys.argv[1:]
if args[0] == 'log_func_call':
    from .tools.add_log_func_call_decorators import main as main_log_func_call
    if len(args) > 1:
        config_path = args[1]
        main_log_func_call(['.'], config_path=config_path)
    else:
        main_log_func_call(['.'])

else:
    raise ValueError(f"Unknown command: {args[0]}")



# This file is auto-generated by pyrandyos/gui/icons/update_from_spec.py  # noqa: E501
# Do not edit.
address_book = 62137
address_card = 62139
angry = 62806
arrow_alt_circle_down = 62296
arrow_alt_circle_left = 62297
arrow_alt_circle_right = 62298
arrow_alt_circle_up = 62299
bell = 61683
bell_slash = 61942
bookmark = 61486
building = 61869
calendar = 61747
calendar_alt = 61555
calendar_check = 62068
calendar_minus = 62066
calendar_plus = 62065
calendar_times = 62067
caret_square_down = 61776
caret_square_left = 61841
caret_square_right = 61778
caret_square_up = 61777
chart_bar = 61568
check_circle = 61528
check_square = 61770
circle = 61713
clipboard = 62248
clock = 61463
clone = 62029
closed_captioning = 61962
comment = 61557
comment_alt = 62074
comment_dots = 62637
comments = 61574
compass = 61774
copy = 61637
copyright = 61945
credit_card = 61597
dizzy = 62823
dot_circle = 61842
edit = 61508
envelope = 61664
envelope_open = 62134
eye = 61550
eye_slash = 61552
file = 61787
file_alt = 61788
file_archive = 61894
file_audio = 61895
file_code = 61897
file_excel = 61891
file_image = 61893
file_pdf = 61889
file_powerpoint = 61892
file_video = 61896
file_word = 61890
flag = 61476
flushed = 62841
folder = 61563
folder_open = 61564
font_awesome_logo_full = 62694
frown = 61721
frown_open = 62842
futbol = 61923
gem = 62373
grimace = 62847
grin = 62848
grin_alt = 62849
grin_beam = 62850
grin_beam_sweat = 62851
grin_hearts = 62852
grin_squint = 62853
grin_squint_tears = 62854
grin_stars = 62855
grin_tears = 62856
grin_tongue = 62857
grin_tongue_squint = 62858
grin_tongue_wink = 62859
grin_wink = 62860
hand_lizard = 62040
hand_paper = 62038
hand_peace = 62043
hand_point_down = 61607
hand_point_left = 61605
hand_point_right = 61604
hand_point_up = 61606
hand_pointer = 62042
hand_rock = 62037
hand_scissors = 62039
hand_spock = 62041
handshake = 62133
hdd = 61600
heart = 61444
hospital = 61688
hourglass = 62036
id_badge = 62145
id_card = 62146
image = 61502
images = 62210
keyboard = 61724
kiss = 62870
kiss_beam = 62871
kiss_wink_heart = 62872
laugh = 62873
laugh_beam = 62874
laugh_squint = 62875
laugh_wink = 62876
lemon = 61588
life_ring = 61901
lightbulb = 61675
list_alt = 61474
map = 62073
meh = 61722
meh_blank = 62884
meh_rolling_eyes = 62885
minus_square = 61766
money_bill_alt = 62417
moon = 61830
newspaper = 61930
object_group = 62023
object_ungroup = 62024
paper_plane = 61912
pause_circle = 62091
play_circle = 61764
plus_square = 61694
question_circle = 61529
registered = 62045
sad_cry = 62899
sad_tear = 62900
save = 61639
share_square = 61773
smile = 61720
smile_beam = 62904
smile_wink = 62682
snowflake = 62172
square = 61640
star = 61445
star_half = 61577
sticky_note = 62025
stop_circle = 62093
sun = 61829
surprise = 62914
thumbs_down = 61797
thumbs_up = 61796
times_circle = 61527
tired = 62920
trash_alt = 62189
user = 61447
user_circle = 62141
window_close = 62480
window_maximize = 62160
window_minimize = 62161
window_restore = 62162



from collections.abc import Iterable
from pathlib import Path
from hashlib import new as newhash
from re import search

from ..logging import log_func_call
from .string import iterable_max_chars
from .tqdm import FileSetTqdm

FileSet = set[Path]
StrTup = tuple[str]
FilePair = tuple[Path, Path]

BULLETSEP = '\n* '


@log_func_call
def filehash(p: Path, *, blocksize: int = 65536,
             algorithm: str = 'md5') -> str:
    hasher = newhash(algorithm)
    with p.open('rb') as f:
        while chunk := f.read(blocksize):
            hasher.update(chunk)
    return hasher.hexdigest()


@log_func_call
def is_file_in_ignore_regex(p: Path, ignore: StrTup = ()):
    for x in ignore:
        if search(x, p.as_posix()):
            return True

    return False


@log_func_call
def is_file_in_ignore_parts(p: Path, ignore: StrTup = ()):
    parts = p.parts
    for x in ignore:
        if x in parts:
            return True

    return False


@log_func_call
def is_file_in_ignore_suffix(p: Path, ignore: StrTup = ()):
    suffix = p.suffix
    joinsuffix = ''.join(p.suffixes)
    for x in ignore:
        if suffix == x or joinsuffix == x:
            return True
    return False


@log_func_call
def is_file_in_ignore_dirs(p: Path, ignore: StrTup = ()):
    for x in ignore:
        if p.is_relative_to(Path(x)):
            return True
    return False


@log_func_call
def is_file_in_fileset(p: Path, files: StrTup = ()):
    return p.as_posix() in files


@log_func_call
def is_file_in_blacklist(p: Path, blacklist: StrTup = ()):
    return is_file_in_fileset(p, blacklist)


@log_func_call
def is_file_in_whitelist(p: Path, whitelist: StrTup = ()):
    return is_file_in_fileset(p, whitelist)


@log_func_call
def should_ignore_file(p: Path,
                       blacklist: StrTup = (),
                       dir_ignores: StrTup = (),
                       parts_ignores: StrTup = (),
                       suffix_ignores: StrTup = (),
                       regex_ignores: StrTup = ()):
    return (is_file_in_blacklist(p, blacklist)
            or is_file_in_ignore_dirs(p, dir_ignores)
            or is_file_in_ignore_parts(p, parts_ignores)
            or is_file_in_ignore_suffix(p, suffix_ignores)
            or is_file_in_ignore_regex(p, regex_ignores))


@log_func_call
def build_ignore_args_dict(blacklist: StrTup = (),
                           dir_ignores: StrTup = (),
                           parts_ignores: StrTup = (),
                           suffix_ignores: StrTup = (),
                           regex_ignores: StrTup = ()):
    return dict(blacklist=blacklist,
                dir_ignores=dir_ignores,
                parts_ignores=parts_ignores,
                suffix_ignores=suffix_ignores,
                regex_ignores=regex_ignores)


@log_func_call
def generate_fileset(p: Path,
                     apply_filter: bool = True,
                     whitelist: StrTup = (),
                     blacklist: StrTup = (),
                     dir_ignores: StrTup = (),
                     parts_ignores: StrTup = (),
                     suffix_ignores: StrTup = (),
                     regex_ignores: StrTup = ()):
    ignore_args = build_ignore_args_dict(blacklist, dir_ignores, parts_ignores,
                                         suffix_ignores, regex_ignores)
    s = set()
    for f in p.rglob('*'):
        if f.is_dir():
            continue
        relpath = f.relative_to(p)
        if (apply_filter
                and not is_file_in_whitelist(relpath, whitelist)
                and should_ignore_file(relpath, **ignore_args)):
            continue

        s.add(relpath)

    return s


@log_func_call
def compare_dirs(src: Path, dest: Path,
                 filter: bool = True):
    return compare_filesets(*get_src_dest_filesets(src, dest, filter))


@log_func_call
def get_src_dest_filesets(src: Path, dest: Path,
                          apply_filter: bool = True,
                          whitelist: StrTup = (),
                          blacklist: StrTup = (),
                          dir_ignores: StrTup = (),
                          parts_ignores: StrTup = (),
                          suffix_ignores: StrTup = (),
                          regex_ignores: StrTup = ()):
    ignore_args = build_ignore_args_dict(blacklist, dir_ignores, parts_ignores,
                                         suffix_ignores, regex_ignores)
    kwargs = dict(whitelist=whitelist, **ignore_args)
    src_files = generate_fileset(src, apply_filter, **kwargs)
    dest_files = generate_fileset(dest, apply_filter, **kwargs)
    return src_files, dest_files


@log_func_call
def compare_filesets(src_files: FileSet, dest_files: FileSet):
    files_src_only = src_files.difference(dest_files)
    files_dest_only = dest_files.difference(src_files)
    files_on_both = src_files.intersection(dest_files)
    return files_src_only, files_dest_only, files_on_both


@log_func_call
def fileset_as_posix(fset: FileSet):
    return {x.as_posix() for x in fset}


@log_func_call
def fileset_max_chars(fset: FileSet):
    return iterable_max_chars(fileset_as_posix(fset))


@log_func_call
def compare_fileset_hashes(fset: FileSet, src: Path, dest: Path,
                           algorithm: str = 'md5', verbose: bool = True):
    not_matching = set()
    for f in FileSetTqdm(fset):
        srcfile = src/f
        srchash = filehash(srcfile, algorithm=algorithm)

        destfile = dest/f
        desthash = filehash(destfile, algorithm=algorithm)

        if desthash != srchash:
            if verbose:
                print(f'{f} | src: {srchash}, dest: {desthash}')

            not_matching.add(f)

    return not_matching


@log_func_call
def fileset_to_sorted_str_list(fset: FileSet):
    return sorted(fileset_as_posix(fset))


@log_func_call
def sorted_bulleted_list(x: Iterable[str], sep: str = BULLETSEP):
    return sep + sep.join(sorted(x))


@log_func_call
def fileset_to_str(fset: FileSet, sep: str = BULLETSEP):
    return sorted_bulleted_list(fileset_to_sorted_str_list(fset), sep)


@log_func_call
def print_fileset(fset: FileSet):
    print(fileset_to_str(fset))


@log_func_call
def src_dest_pairs(src: Path, dest: Path, fset: FileSet):
    return {(src/f, dest/f) for f in fset}


@log_func_call
def generate_md5sum_file(fset: FileSet, md5file: Path = None,
                         base_path: Path = None):
    out = ''
    base_path = base_path or (md5file.parent if md5file else Path.cwd())
    relroot = md5file.parent if md5file else base_path
    for f in FileSetTqdm(fset, desc='Generating md5sum file'):
        p = f if f.is_absolute() else base_path/f
        md5 = filehash(p, algorithm='md5')  # hardcoding for `md5sum` compat
        out += f'{md5}  {p.relative_to(relroot).as_posix()}\n'

    if md5file:
        md5file.write_text(out)

    return out


@log_func_call
def parse_md5sum_file(md5file: Path, base_path: Path = None):
    base_path = base_path or md5file.parent
    return parse_md5sum_file_text(md5file.read_text(), base_path)


@log_func_call
def parse_md5sum_file_text(md5text: str, base_path: Path = None):
    base_path = base_path or Path.cwd()
    md5data: dict[Path, str] = dict()
    for line in md5text.splitlines():
        md5, f = line.split('  ', 1)
        f = Path(f)
        if not f.is_absolute():
            f = base_path/f

        md5data[f] = md5

    return md5data


@log_func_call
def check_md5sum_file(md5file: Path, base_path: Path = None,
                      verbose: bool = True):
    base_path = base_path or md5file.parent
    md5data = parse_md5sum_file(md5file, base_path)
    fset = set(md5data.keys())
    not_matching = set()
    for f in FileSetTqdm(fset, desc='Checking md5sum file'):
        theirhash = md5data[f]
        ourhash = filehash(f, algorithm='md5')
        if theirhash != ourhash:
            if verbose:
                print(f'{f} | ours: {ourhash}, theirs: {theirhash}')

            not_matching.add(f)

    return not_matching



from pandas import Series
from numpy import divmod as npdivmod

from ..logging import log_func_call


@log_func_call
def round_half_away(value: float | int | str | Series | None = None,
                    ndigits: int = 0):
    if not isinstance(value, Series):
        value = float(value)

    sgn = 1 - 2*(value < 0)
    absval = abs(value)
    expo = -ndigits or 0
    scalar = 10**expo
    quo, rem = npdivmod(absval, scalar)
    out = sgn*scalar*(quo + (rem*2 >= scalar))
    if ndigits <= 0:
        return out.astype(int) if isinstance(out, Series) else int(out)
    return out



import sys as _sys
from os import environ as _environ

from typing import (
    TypeVar as _TypeVar, overload as _overload, Any as _Any
)
from types import TracebackType as _TracebackType
from collections.abc import Callable as _Callable, Mapping as _Mapping
from traceback import format_exception_only
from functools import partial as _partial
# want to export these for convenience, so they are not hidden by default
from logging import (  # noqa: F401
    getLogger as _getLogger, WARN, ERROR, DEBUG, INFO, CRITICAL, WARNING,
    Logger, getLevelName as _getLevelName, LogRecord as _LogRecord,
)
try:
    from logging import getLevelNamesMapping as _getLevelNamesMapping
except ImportError:
    # not available until Python 3.12
    _getLevelNamesMapping = None

from .utils.signature_wrapper import (
    generate_signature_aware_wrapper as _sig_aware_wrapper
)
from ._testing.debug import is_debug_enabled
from .utils.stack import (
    exc_info as _exc_info, log_find_caller as _find_caller,
    format_exc as _format_exc, get_stack_frame as _get_stack_frame,
    ExcInfoType as _ExcInfoType,
    is_code_the_given_func as _is_code_the_given_func,
)

_RecFactoryType = _Callable[..., _LogRecord]


Logger.root.setLevel(0)

LOGSTDOUT = INFO + 1
LOGSTDERR = INFO + 2
LOGTQDM = INFO + 3
DEBUGLOW = DEBUG - 1
DEBUGLOW2 = DEBUG - 2

APP_LOG_LEVEL_NAMES = {
    'STDOUT': LOGSTDOUT,
    'STDERR': LOGSTDERR,
    'TQDM': LOGTQDM,
    'DEBUGLOW': DEBUGLOW,
    'DEBUGLOW2': DEBUGLOW2,
}

TRACELOG: bool = False
FUNCCALLLOG: bool = False
_GLOBAL_LOG: Logger = None


def get_global_logger():
    return _GLOBAL_LOG


def set_global_logger(log: Logger):
    global _GLOBAL_LOG
    _GLOBAL_LOG = log


def get_logger(modname: str = None, stacklevel: int = 2) -> Logger:
    __traceback_hide__ = True  # noqa: F841
    log = _GLOBAL_LOG
    if log:
        return log
    # default stack level here is 2 because we want to pop off both
    # get_stack_frame and get_logger.
    if not modname:
        f = _get_stack_frame(stacklevel)
        if f:
            modname = f.f_globals['__name__']

        else:
            # if we exhausted the whole stack, it's probably because the issue
            # is in the main script, so let's just assume it's '__main__'
            modname = '__main__'

    return _getLogger(modname)


def log_message(level: int | str, msg: str, *args,
                exc_info: _ExcInfoType | BaseException = None,
                extra: _Mapping[str, object] = None, stack_info: bool = False,
                stacklevel: int = 1):
    # the default stacklevel is 1 because we assume we don't want to include
    # the call to log_message, which would be first on the stack after removing
    # all the internals already.
    #
    # the stacklevel for get_logger, however, is an ABSOLUTE stack level
    # since the value is passed directly to get_stack_frame.  Therefore,
    # the value we pass here needs to account for:
    # 1. get_stack_frame
    # 2. get_logger
    # 3. log_message
    # This is why the offset is two because we assume stacklevel is starting to
    # count up the stack including log_message, so we only need the extra
    # levels for get_logger and get_stack_frame.
    #
    # Finally, the stacklevel for _log itself only adds one level because the
    # stacklevel there is used to pop elements off the end of the stack list
    # from filter_stack.  With a stacklevel of 1, it should return the first
    # non-hidden function it finds (because it's actually a -1).  However,
    # THIS function is NOT hidden with __traceback_hide__, and therefore this
    # is the function it would return.  Since we assume we actually want the
    # caller of THIS function, we need to pop one additional level.
    _log(get_logger(stacklevel=stacklevel + 2), level, msg, *args,
         exc_info=exc_info, extra=extra, stack_info=stack_info,
         stacklevel=stacklevel + 1)


log_debuglow2 = _partial(log_message, DEBUGLOW2)
log_debuglow = _partial(log_message, DEBUGLOW)
log_debug = _partial(log_message, DEBUG)
log_info = _partial(log_message, INFO)
log_warning = _partial(log_message, WARNING)
log_error = _partial(log_message, ERROR)
log_critical = _partial(log_message, CRITICAL)


def _log_func_call_handler(handler_args: tuple, handler_kwargs: dict,
                           func: _Callable, *func_args: tuple,
                           **func_kwargs: dict):
    __traceback_hide__ = True  # noqa: F841
    level: str | int = handler_args[0]
    trace_only: bool = handler_kwargs.get('trace_only', False)
    stacklevel: int = handler_kwargs.get('stacklevel', 1)
    if get_func_call_logging() and (not trace_only or get_tracelog()):
        log = get_logger(func.__module__)
        try:
            code = func.__code__
            funcname = func.__name__
            funcargstr = ', '.join(
                'self' if not i and funcname == '__init__'
                else repr(arg) for i, arg in enumerate(func_args)
            )
            funcargstr += ', ' if func_args and func_kwargs else ''
            funcargstr += ', '.join(f'{k}={v!r}'
                                    for k, v in func_kwargs.items())
            _log(
                log,
                level,
                f"{'TRACE: ' if trace_only else ''}"
                f"Function call: {func.__qualname__}"
                f"({funcargstr}) "
                f"{{function defined {code.co_filename}"
                f"({code.co_firstlineno})}}",
                stacklevel=stacklevel,
            )
        except BaseException as e:
            _log(
                log,
                level,
                "Error logging function call: "
                f"{func.__qualname__} - "
                f"{''.join(format_exception_only(e)).strip()} "
                f"{{function defined {code.co_filename}"
                f"({code.co_firstlineno})}}",
                stacklevel=stacklevel,
            )

    return func_args, func_kwargs


F = _TypeVar("F", bound=_Callable[..., _Any])
@_overload
def log_func_call(func: F) -> F: ...
@_overload  # noqa: E302
def log_func_call(level: int | str, *,
                  trace_only: bool = False,
                  stacklevel: int = 1) -> _Callable[[F], F]: ...
def log_func_call(arg, *,  # noqa: E302
                  trace_only: bool = False,
                  stacklevel: int = 1):
    if _environ.get("PYRANDYOS_BYPASS_CALL_LOG"):
        return arg if callable(arg) else lambda f: f

    def log_decorator(func: F) -> F:
        return _sig_aware_wrapper(func, _log_func_call_handler, level,
                                  trace_only=trace_only, stacklevel=stacklevel)

    if callable(arg):
        # Used as @log_func_call
        level = DEBUGLOW
        return log_decorator(arg)
    else:
        # Used as @log_func_call(level)
        level = arg
        return log_decorator


def set_func_call_logging(enabled: bool = True):
    global FUNCCALLLOG
    FUNCCALLLOG = bool(enabled)


def get_func_call_logging() -> bool:
    global FUNCCALLLOG
    return FUNCCALLLOG


def set_trace_logging(enabled: bool = True):
    global TRACELOG
    TRACELOG = bool(enabled)


def get_tracelog() -> bool:
    global TRACELOG
    return TRACELOG


def log_exc(exc_or_type: type | BaseException = None,
            exc: BaseException = None,
            traceback: _TracebackType = None,
            msg: str = 'Unhandled exception',
            mark_handled: bool = True,
            stacklevel: int = 1):
    __traceback_hide__ = True  # noqa: F841
    # see comments on stacklevel in log_message.  Note that this function IS
    # hidden, so no offset is applied for the arg to _log.
    excnfo = _exc_info(exc_or_type, exc, traceback)
    _log(get_logger(stacklevel=stacklevel + 2), ERROR, msg, exc_info=excnfo,
         stacklevel=stacklevel)
    if mark_handled:
        excnfo[1]._pyrandyos_handled = True


def _log_exc_hook(exc_or_type: type | BaseException = None,
                  exc: BaseException = None,
                  traceback: _TracebackType = None):
    f = _get_stack_frame(2)  # excludes _log_exc_hook and get_stack_frame
    # if we are in debug, usually we want to have the debugger handle
    # exceptions so that we can break on them.  If we log the exception with
    # log_exc AND pass to the sys.__excepthook__, it would log the exception
    # twice (at least to the console, perhaps not the log file).  This is why
    # we only enable log_exc here when we are NOT debugging, because otherwise
    # the sys.__excepthook__ would not record the error in the log and would
    # die instantly.  We have multiple mechanisms already in PyRandyOS to try
    # to catch and log exceptions, but if one is showing up here, it somehow
    # got through the cracks since this is the handler of last resort.
    #
    # It seems that the way you can end up here is if exceptions are raised
    # from outside the Python interpreter, like during a callback invoked from
    # a foreign function via ctypes or Qt for example.  QtApp.notify is the
    # event handler for the GUI event loop, so we can check for that as the
    # source of the exception that might otherwise not get caught by the usual
    # Python mechanisms.  However, it is not the only one even with Qt.
    # The signals/slots paradigm for other non-event based callbacks are
    # instead invoked directly and thus do not pass through QtApp.notify.
    # This means there is otherwise not a reliable way to detect those since
    # the Qt caller will be invisible in the Python call stack for all intents
    # and purposes.  We have to rely on the user to decorate slots so that we
    # can know definitively they are being called externally and thus worthy
    # of trapping here even in a debug scenario.
    tbf = traceback.tb_frame
    code = tbf.f_code
    from .gui.callback import QtCallable
    is_qt_callback = _is_code_the_given_func(QtCallable.__call__, code)

    debug = is_debug_enabled()
    is_qt_notify = False
    if debug and f:
        code = f.f_code
        from .gui.gui_app import QtApp
        is_qt_notify = _is_code_the_given_func(QtApp.notify, code)

    if (not debug or not f or is_qt_notify or is_qt_callback):
        # stacklevel here needs to pop an additional level from the default 1
        # for the absolute stacklevel of _log_exc_hook itself for get_logger.
        # The secondary use of stacklevel otherwise would be for getting the
        # right output from filter_stack, but that is only called when not
        # given exc_info.  since we are passing exc_info, we only need to
        # cover the get_logger use case.
        log_exc(exc_or_type, exc, traceback, stacklevel=2)

    if not getattr(exc, '_pyrandyos_handled', False):
        # if we did not handle it, let the default handler do its job
        _sys.__excepthook__(exc_or_type, exc, traceback)


_sys.excepthook = _log_exc_hook
# _sys.excepthook = (lambda: None) if is_debug_enabled() else log_exc


def _default_rec_factory(name: str, level: int, fn: str, lno: int, msg,
                         args: tuple, exc_info: _ExcInfoType, func: str = None,
                         extra: dict = None, sinfo: str = None):
    __traceback_hide__ = True  # noqa: F841
    # adapted from Logger.makeRecord
    rv = _LogRecord(name, level, fn, lno, msg, args, exc_info, func, sinfo)
    if extra is not None:
        for key in extra:
            if (key in ["message", "asctime"]) or (key in rv.__dict__):
                raise KeyError("Attempt to overwrite %r in LogRecord" % key)
            rv.__dict__[key] = extra[key]
    return rv


def make_log_record(level: int | str, msg: str, *args,
                    exc_info: _ExcInfoType | BaseException = None,
                    extra: _Mapping[str, object] = None,
                    stack_info: bool = False, stacklevel: int = 1,
                    logname: str | None = None,
                    recfactory: _RecFactoryType = _default_rec_factory):
    __traceback_hide__ = True  # noqa: F841
    if not isinstance(level, int):
        level, _ = get_loglevel_num_name(level)

    if logname is None:
        # this stacklevel does NOT exclude __traceback_hide__ functions
        # so we must adjust for it.  If we are calling this function
        # (make_log_record) directly, we assume the default stacklevel 1
        # accounts for this frame, the +2 is the same as in log_message,
        # and then we end up at the caller of make_log_record, which is the
        # desired outcome in this case.
        logname = get_logger(stacklevel=stacklevel + 2).name

    exc_info = _exc_info(exc_info, skip_if_none=True)
    # no need to adjust stacklevel because it excludes __traceback_hide__
    # functions in the stack automatically here, including this one, so it
    # should return the caller of this function (or any non-hidden predecessor)
    fn, lno, func, sinfo = _find_caller(stack_info, stacklevel, exc_info)
    record = recfactory(logname, level, fn, lno, msg, args, exc_info,
                        func, extra, sinfo)
    if exc_info:
        record.exc_text = _format_exc(exc_info[1], exc_info[2])

    return record


def _log(log: Logger, level: int | str, msg: str, *args,
         exc_info: _ExcInfoType | BaseException = None,
         extra: _Mapping[str, object] = None,
         stack_info: bool = False, stacklevel: int = 1):
    __traceback_hide__ = True  # noqa: F841
    # no need to adjust stacklevel because it excludes __traceback_hide__
    # functions in the stack automatically
    log.handle(make_log_record(level, msg, *args,
                               exc_info=exc_info, extra=extra,
                               stack_info=stack_info,
                               stacklevel=stacklevel,
                               logname=log.name,
                               recfactory=log.makeRecord))


# @log_func_call(DEBUGLOW2)
def log_level_by_name(name: str):
    if _getLevelNamesMapping:
        return _getLevelNamesMapping().get(name)
    # NOTE: the logic in here is apparently a bug and deprecated, but isn't
    # fixed until Python 3.12.  Supporting both so I don't have to change
    # this code later once we upgrade Python versions.
    num = _getLevelName(name)
    if isinstance(num, str):
        return
    return num


# @log_func_call(DEBUGLOW2)
def get_loglevel_num_name(level: str | int):
    "returns num, name"
    if isinstance(level, str):
        level = level.upper()
        num = log_level_by_name(level)
        if num is None:
            raise ValueError(f'unknown loglevel string {level} given')
        return num, level
    name = _getLevelName(level)
    num = log_level_by_name(name)
    if not isinstance(num, int):
        name = None

    return level, name



"""
Compute the version number and store it in the `__version__` variable.

Based on <https://github.com/maresb/hatch-vcs-footgun-example>.
"""


def _get_hatch_version():
    """
    Compute the most up-to-date version number in a development environment.

    Returns `None` if Hatchling is not installed, e.g. in a prod environment.

    For more details, see:
    <https://github.com/maresb/hatch-vcs-footgun-example/>.
    """
    import os

    try:
        from hatchling.metadata.core import ProjectMetadata
        from hatchling.plugin.manager import PluginManager
        from hatchling.utils.fs import locate_file
        from hatchling.plugin.exceptions import UnknownPluginError
    except ImportError:
        # Hatchling is not installed, so probably we are not in
        # a development environment.
        return None

    pyproject_toml = locate_file(__file__, "pyproject.toml")
    if pyproject_toml:
        root = os.path.dirname(pyproject_toml)
        metadata = ProjectMetadata(root=root, plugin_manager=PluginManager())
        # can be either statically set in pyproject.toml
        # or computed dynamically:
        try:
            return metadata.core.version or metadata.hatch.version.cached
        except UnknownPluginError:
            return None

    # pyproject.toml not found although hatchling is installed
    return None


def _get_importlib_metadata_version():
    """Compute the version number using importlib.metadata.

    This is the official Pythonic way to get the version number of an installed
    package. However, it is only updated when a package is installed. Thus, if
    a package is installed in editable mode, and a different version is checked
    out, then the version number will not be updated.
    """
    from importlib.metadata import version, PackageNotFoundError

    try:
        __version__ = version(__package__)
    except PackageNotFoundError:
        # raise
        __version__ = None
    return __version__


__dynamic_version__ = _get_hatch_version() or _get_importlib_metadata_version()

try:
    from ._version import __version__  # pyright: ignore[reportMissingImports]
except (ModuleNotFoundError, ImportError):
    __version__ = __dynamic_version__



import sys
from inspect import signature
from collections.abc import Callable
from functools import partial, update_wrapper
from textwrap import dedent

from .stack import (
    safe_exec, get_framesummary_for_frame, get_stack_frame,
    mark_next_tb_reraise_to_skip,
)


def generate_signature_aware_wrapper(func: Callable, arg_handler: Callable,
                                     *handler_args, **handler_kwargs):
    # __traceback_hide__ = True  # noqa: F841
    __traceback_hide_locals__ = True  # noqa: F841

    # If func is a partial, get the original function
    # to ensure signature and metadata are correct.
    if isinstance(func, partial):
        original_func = func.func
    else:
        original_func = func

    handler_partial = partial(arg_handler, handler_args, handler_kwargs, func)
    sig = signature(func)
    sigparams = sig.parameters

    # Build code to collect args and kwargs from the named parameters
    args_list = list()
    kwargs_list = list()
    param_list = list()
    defaults = dict()
    for p in sigparams.values():
        if p.kind in (p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD):
            args_list.append(p.name)
            if p.default is not p.empty:
                param_list.append(f'{p.name}=__defaults[{p.name!r}]')
                defaults[p.name] = p.default
            else:
                param_list.append(p.name)
        elif p.kind == p.VAR_POSITIONAL:
            s = f'*{p.name}'
            args_list.append(s)
            param_list.append(s)
        elif p.kind == p.KEYWORD_ONLY:
            kwargs_list.append(f'{p.name}={p.name}')
            if p.default is not p.empty:
                param_list.append(f'{p.name}=__defaults[{p.name!r}]')
                defaults[p.name] = p.default
            else:
                param_list.append(p.name)
        elif p.kind == p.VAR_KEYWORD:
            s = f'**{p.name}'
            kwargs_list.append(s)
            param_list.append(s)

    args_str = ', '.join(args_list)
    kwargs_str = ', '.join(kwargs_list)
    params = ', '.join(param_list)
    bindparams = f'{args_str}{", " if kwargs_str else ""}{kwargs_str}'

    # Create the wrapper source code with the same signature
    f = get_stack_frame()
    src_frame = get_framesummary_for_frame(f, lineno=f.f_lineno + 1)
    src = dedent(f"""
    def signature_aware_wrapper({params}):
        __traceback_hide__ = True  # noqa: F841
        try:
            __bound = __sig.bind({bindparams})
            __bound.apply_defaults()
            __func_args, __func_kwargs = __handler_partial(*__bound.args,
                                                           **__bound.kwargs)
            return __func(*__func_args, **__func_kwargs)
        except BaseException as __e:
            __e._pyrandyos_exec_source = __src
            __e._pyrandyos_exec_source_frame = __src_frame
            __mark_next_tb_reraise_to_skip(__e)
            raise __e
    """)

    # Create wrapper in a safe namespace
    env = {'__func': func, '__sig': sig, '__handler_partial': handler_partial,
           '__src': src, '__src_frame': src_frame, '__defaults': defaults,
           '__mark_next_tb_reraise_to_skip': mark_next_tb_reraise_to_skip}
    loc = None  # {'func_args': None, 'func_kwargs': None,}
    e_to_raise = None
    modules = sys.modules
    tryagain = True
    while tryagain:
        tryagain = False
        try:
            safe_exec(src, env, loc, log_errors=False, src_frame=src_frame)
        except NameError as e:
            name = e.name
            if name in modules:
                # If the name is in sys.modules, it might be a module import
                # that needs to be added to the environment.
                env[name] = modules[name]
                tryagain = True

            else:
                # If the name is not found, it might be a real missing import
                # or other name error or typo.
                # Raise the original exception.
                e_to_raise = e

        except BaseException as e:
            e_to_raise = e

    if e_to_raise:
        try:
            from ..logging import log_exc
        except ImportError:
            pass
        else:
            log_exc(e_to_raise)

        raise e_to_raise

    signature_aware_wrapper = env['signature_aware_wrapper']
    update_wrapper(signature_aware_wrapper, func)

    # Copy metadata (from the original function if it's a partial)
    signature_aware_wrapper.__name__ = original_func.__name__
    signature_aware_wrapper.__doc__ = original_func.__doc__
    signature_aware_wrapper.__module__ = original_func.__module__

    return signature_aware_wrapper


def example_decorator_arg_handler(handler_args: tuple, handler_kwargs: dict,
                                  func: Callable, *func_args, **func_kwargs):
    """
    This method can be used to handle the function arguments
    before calling the actual function.  Override this method
    in subclasses to customize argument handling.
    """
    print(f"Decorator args: {handler_args}, kwargs: {handler_kwargs}")
    print(f"Function args: {func_args}, kwargs: {func_kwargs}")
    print(f"Function name: {func.__qualname__}")
    return func_args, func_kwargs


def example_decorator(message: str):
    def example_decorator_factory(func):
        return generate_signature_aware_wrapper(func,
                                                example_decorator_arg_handler,
                                                message)
    return example_decorator_factory


@example_decorator("argument passed to decorator at compile time, "
                   "handled by arg_handler at run time")
def example_decorated_function():
    print("Example function executed.")



# This file is auto-generated by pyrandyos/gui/icons/update_from_spec.py  # noqa: E501
# Do not edit.
address_book = 61698
address_book_alt = 61697
adjust = 61700
adjust_alt = 61699
adult = 61701
align_center = 61702
align_justify = 61703
align_left = 61704
align_right = 61705
arrow_down = 61706
arrow_left = 61707
arrow_right = 61708
arrow_up = 61709
asl = 61710
asterisk = 61711
backward = 61712
ban_circle = 61713
barcode = 61714
behance = 61715
bell = 61716
blind = 61717
blogger = 61718
bold = 61719
book = 61720
bookmark = 61722
bookmark_empty = 61721
braille = 61723
briefcase = 61724
broom = 61725
brush = 61726
bulb = 61727
bullhorn = 61728
calendar = 61730
calendar_sign = 61729
camera = 61731
car = 61732
caret_down = 61733
caret_left = 61734
caret_right = 61735
caret_up = 61736
cc = 61737
certificate = 61738
check = 61740
check_empty = 61739
chevron_down = 61741
chevron_left = 61742
chevron_right = 61743
chevron_up = 61744
child = 61745
circle_arrow_down = 61746
circle_arrow_left = 61747
circle_arrow_right = 61748
circle_arrow_up = 61749
cloud = 61751
cloud_alt = 61750
cog = 61753
cog_alt = 61752
cogs = 61754
comment = 61756
comment_alt = 61755
compass = 61758
compass_alt = 61757
credit_card = 61759
css = 61760
dashboard = 61761
delicious = 61762
deviantart = 61763
digg = 61764
download = 61766
download_alt = 61765
dribbble = 61767
edit = 61768
eject = 61769
envelope = 61771
envelope_alt = 61770
error = 61773
error_alt = 61772
eur = 61774
exclamation_sign = 61775
eye_close = 61776
eye_open = 61777
facebook = 61778
facetime_video = 61779
fast_backward = 61780
fast_forward = 61781
female = 61782
file = 61788
file_alt = 61783
file_edit = 61785
file_edit_alt = 61784
file_new = 61787
file_new_alt = 61786
film = 61789
filter = 61790
fire = 61791
flag = 61793
flag_alt = 61792
flickr = 61794
folder = 61798
folder_close = 61795
folder_open = 61796
folder_sign = 61797
font = 61799
fontsize = 61800
fork = 61801
forward = 61803
forward_alt = 61802
foursquare = 61804
friendfeed = 61806
friendfeed_rect = 61805
fullscreen = 61807
gbp = 61808
gift = 61809
github = 61811
github_text = 61810
glass = 61812
glasses = 61813
globe = 61815
globe_alt = 61814
googleplus = 61816
graph = 61818
graph_alt = 61817
group = 61820
group_alt = 61819
guidedog = 61821
hand_down = 61822
hand_left = 61823
hand_right = 61824
hand_up = 61825
hdd = 61826
headphones = 61827
hearing_impaired = 61828
heart = 61831
heart_alt = 61829
heart_empty = 61830
home = 61833
home_alt = 61832
hourglass = 61834
idea = 61836
idea_alt = 61835
inbox = 61839
inbox_alt = 61837
inbox_box = 61838
indent_left = 61840
indent_right = 61841
info_circle = 61842
instagram = 61843
iphone_home = 61844
italic = 61845
key = 61846
laptop = 61848
laptop_alt = 61847
lastfm = 61849
leaf = 61850
lines = 61851
link = 61852
linkedin = 61853
list = 61855
list_alt = 61854
livejournal = 61856
lock = 61858
lock_alt = 61857
magic = 61859
magnet = 61860
male = 61861
map_marker = 61863
map_marker_alt = 61862
mic = 61865
mic_alt = 61864
minus = 61867
minus_sign = 61866
move = 61868
music = 61869
myspace = 61870
network = 61871
off = 61872
ok = 61875
ok_circle = 61873
ok_sign = 61874
opensource = 61876
paper_clip = 61878
paper_clip_alt = 61877
path = 61879
pause = 61881
pause_alt = 61880
pencil = 61883
pencil_alt = 61882
person = 61884
phone = 61886
phone_alt = 61885
photo = 61888
photo_alt = 61887
picasa = 61889
picture = 61890
pinterest = 61891
plane = 61892
play = 61895
play_alt = 61893
play_circle = 61894
plurk = 61897
plurk_alt = 61896
plus = 61899
plus_sign = 61898
podcast = 61900
print = 61901
puzzle = 61902
qrcode = 61903
question = 61905
question_sign = 61904
quote_alt = 61906
quote_right = 61908
quote_right_alt = 61907
quotes = 61909
random = 61910
record = 61911
reddit = 61912
redux = 61913
refresh = 61914
remove = 61917
remove_circle = 61915
remove_sign = 61916
repeat = 61919
repeat_alt = 61918
resize_full = 61920
resize_horizontal = 61921
resize_small = 61922
resize_vertical = 61923
return_key = 61924
retweet = 61925
reverse_alt = 61926
road = 61927
rss = 61928
scissors = 61929
screen = 61931
screen_alt = 61930
screenshot = 61932
search = 61934
search_alt = 61933
share = 61936
share_alt = 61935
shopping_cart = 61938
shopping_cart_sign = 61937
signal = 61939
skype = 61940
slideshare = 61941
smiley = 61943
smiley_alt = 61942
soundcloud = 61944
speaker = 61945
spotify = 61946
stackoverflow = 61947
star = 61950
star_alt = 61948
star_empty = 61949
step_backward = 61951
step_forward = 61952
stop = 61954
stop_alt = 61953
stumbleupon = 61955
tag = 61956
tags = 61957
tasks = 61958
text_height = 61959
text_width = 61960
th = 61963
th_large = 61961
th_list = 61962
thumbs_down = 61964
thumbs_up = 61965
time = 61967
time_alt = 61966
tint = 61968
torso = 61969
trash = 61971
trash_alt = 61970
tumblr = 61972
twitter = 61973
universal_access = 61974
unlock = 61976
unlock_alt = 61975
upload = 61977
usd = 61978
user = 61979
viadeo = 61980
video = 61983
video_alt = 61981
video_chat = 61982
view_mode = 61984
vimeo = 61985
vkontakte = 61986
volume_down = 61987
volume_off = 61988
volume_up = 61989
w3c = 61990
warning_sign = 61991
website = 61993
website_alt = 61992
wheelchair = 61994
wordpress = 61995
wrench = 61997
wrench_alt = 61996
youtube = 61998
zoom_in = 61999
zoom_out = 62000



from typing import overload, Any, TypeVar, TYPE_CHECKING
from collections.abc import Callable
from contextlib import contextmanager
from functools import wraps

if TYPE_CHECKING:
    from .splash import GuiSplashScreen

LOAD_STEP_REGISTRY: set[str] = set()
COMPLETED_LOAD_STEPS: set[str] = set()
STARTED_LOAD_STEPS: set[str] = set()

F = TypeVar("F", bound=Callable[..., Any])


def register_load_step(arg: str | Callable):  # noqa: E302
    if callable(arg):
        arg()
    else:
        reg = LOAD_STEP_REGISTRY
        reg.add(arg)


@overload
def register_as_load_step(step_name: str) -> F: ...
@overload
def register_as_load_step(func: Callable) -> F: ...
def register_as_load_step(arg):  # noqa: E302
    register_load_step(arg)

    def register_load_step_decorator(func: F) -> F:
        @wraps(func)
        def register_load_step_wrapper(*args, **kwargs):
            __traceback_hide__ = True  # noqa: F841
            return func(*args, **kwargs)
        return register_load_step_wrapper
    return register_load_step_decorator


def load_status_step(step_name: str, show_step_start: bool = True,
                     show_step_done: bool = False):
    def load_status_step_decorator(func: F) -> F:
        @wraps(func)
        def load_status_step_wrapper(*args, **kwargs):
            __traceback_hide__ = True  # noqa: F841
            with loading_step_context(step_name,
                                      show_step_start=show_step_start,
                                      show_step_done=show_step_done):
                return func(*args, **kwargs)

        register_load_step(step_name)
        return load_status_step_wrapper  # type: ignore
    return load_status_step_decorator


def mark_load_step_completed(step_name: str, show_step_name: bool = False,
                             process_events: bool = True):
    if step_name in LOAD_STEP_REGISTRY:
        reg = COMPLETED_LOAD_STEPS
        reg.add(step_name)
        from .gui_app import get_gui_app
        splash: GuiSplashScreen = getattr(get_gui_app(), 'splash', None)
        if splash:
            splash.set_progress(value=len(reg),
                                message=step_name if show_step_name else None,
                                process_events=process_events)


def mark_load_step_started(step_name: str, show_step_name: bool = True,
                           process_events: bool = True):
    if step_name in LOAD_STEP_REGISTRY:
        reg = STARTED_LOAD_STEPS
        reg.add(step_name)
        from .gui_app import get_gui_app
        splash: GuiSplashScreen = getattr(get_gui_app(), 'splash', None)
        if splash:
            splash.set_progress(message=step_name if show_step_name else None,
                                process_events=process_events)


def splash_message(message: str, process_events: bool = True):
    from .gui_app import get_gui_app
    splash: GuiSplashScreen = getattr(get_gui_app(), 'splash', None)
    if splash:
        splash.set_progress(message=message, process_events=process_events)


@contextmanager
def loading_step_context(step_name: str, show_step_start: bool = True,
                         show_step_done: bool = False):
    __traceback_hide__ = True  # noqa: F841
    mark_load_step_started(step_name, show_step_start)
    yield
    # don't do a try block because if the block fails, we didn't complete
    mark_load_step_completed(step_name, show_step_done)



from pathlib import Path
from copy import deepcopy

from ..utils.json import load_jsonc, save_json, jsonify
from ..utils.cfgdict import (
    config_dict_update, config_dict_get, config_dict_set,
)
from ..utils.constants import IS_WIN32
from ..utils.casesafe import casesafe_value_in_container, casesafe_is_equal

from ..logging import log_func_call
from .appconfig import AppConfig
from .expandutils import expand_key_recursively
from .keys import get_path_keys, LOCAL_CONFIG_FILE_KEY, LOCAL_CFG_KEY


@log_func_call
def load_local_config():
    local_cfg_path: Path = AppConfig[LOCAL_CONFIG_FILE_KEY]
    if local_cfg_path.exists():
        return load_jsonc(local_cfg_path)
    return dict()


@log_func_call
def process_local_config(base: dict = None, app_path_keys: tuple[str] = (),
                         case_insensitive: bool = None):
    "returns True if a local config was found and loaded, else False"
    if not base:
        base = AppConfig.get_global_config()
        case = AppConfig.get_case(case_insensitive)
    else:
        case = IS_WIN32 if case_insensitive is None else case_insensitive

    local_cfg = load_local_config()
    use_local = bool(local_cfg)
    if use_local:
        from ..logging import log_info
        log_info(f"Using local config: {AppConfig[LOCAL_CONFIG_FILE_KEY]}")

    for k in local_cfg.keys():
        expand_key_recursively(local_cfg, k, case_insensitive=case)

    config_dict_update(get_local_config(base), local_cfg, case)
    AppConfig.process_config(config=base, app_path_keys=app_path_keys,
                             case_insensitive=case)

    return use_local


@log_func_call
def get_local_config(base: dict = None, case_insensitive: bool = None):
    if not base:
        base = AppConfig.get_global_config()
        case = AppConfig.get_case(case_insensitive)
    else:
        case = IS_WIN32 if case_insensitive is None else case_insensitive

    local_cfg: dict = config_dict_get(base, LOCAL_CFG_KEY, {}, case)
    return local_cfg


@log_func_call
def save_local_config(app_path_keys: tuple = (),
                      case_insensitive: bool = None):
    case = AppConfig.get_case(case_insensitive)

    # get the verbatim contents of "old" local config
    old_local_cfg = load_local_config()

    # make a temp copy of current global config
    cfgtmp = deepcopy(AppConfig.get_global_config())

    # process the "old" local config with the temp global copy to get
    # "old" expansions
    process_local_config(cfgtmp, app_path_keys, case)
    old_local_expanded = get_local_config(cfgtmp, case)

    # check if each key is present in old config.
    # If not, add it to the output.
    # If it is, check that the "expanded" values match.
    # If they match, keep the original unexpanded value.
    # Otherwise, use the new expanded value for lack of better guidance.
    pathkeys = get_path_keys(app_path_keys)
    local_cfg: dict = get_local_config(case_insensitive=case)
    out = old_local_cfg.copy()
    for k, v in local_cfg.items():
        if not casesafe_value_in_container(old_local_cfg, k, case):
            if casesafe_value_in_container(pathkeys, f'{LOCAL_CFG_KEY}.{k}'):
                v: Path
                v = v.as_posix()

            config_dict_set(out, k, v, case)
        else:
            v_old = config_dict_get(old_local_cfg, k, case_insensitive=case)
            v_test = v_old
            if casesafe_value_in_container(pathkeys, f'{LOCAL_CFG_KEY}.{k}'):
                v_test: Path = config_dict_get(old_local_expanded, k,
                                               case_insensitive=case)
                v_test = v_test.as_posix()
                v: Path
                v = v.as_posix()

            config_dict_set(out, k,
                            v if not casesafe_is_equal(v_test, v) else v_old,
                            case)

    # save the new output local config
    local_cfg_path: Path = AppConfig[LOCAL_CONFIG_FILE_KEY]
    save_json(local_cfg_path, jsonify(out))
    from ..logging import log_info
    log_info(f'local config saved to {local_cfg_path}')



from pathlib import Path

from ....app import PyRandyOSApp
from ....logging import log_func_call
from ....utils.json import save_json, jsonify
from ...widgets import GuiWindowLikeParentType
from ...qt import QFileDialog
from .. import GuiDialog
from .view import ConfigTreeDialogView


class ConfigTreeDialog(GuiDialog[ConfigTreeDialogView]):
    @log_func_call
    def __init__(self, gui_parent: GuiWindowLikeParentType):
        super().__init__("Current Configuration", gui_parent)

    @log_func_call
    def get_config(self):
        return PyRandyOSApp.get_global_config()

    @log_func_call
    def show(self):
        # Set dialog as child of parent window to prevent focus issues
        # parent_window = self.gui_parent.gui_view.qtobj
        dialog = self.gui_view.qtobj
        # dialog.setParent(parent_window, dialog.windowFlags())

        # # Ensure proper window attributes
        # dialog.setAttribute(Qt.WA_ShowWithoutActivating, False)
        # dialog.setAttribute(Qt.WA_DeleteOnClose, False)

        # Show and raise the dialog
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()

    @log_func_call
    def create_gui_view(self, basetitle: str, *args,
                        **kwargs) -> ConfigTreeDialogView:
        return ConfigTreeDialogView(basetitle, self, *args, **kwargs)

    @log_func_call
    def click_save_config(self):
        qtwin = self.gui_view.qtobj
        workdir: Path = PyRandyOSApp['tmp_dir']
        if not workdir.exists():
            workdir = Path.cwd()

        new_fn, selected_filter = QFileDialog.getSaveFileName(
            qtwin, "Save Local Config", str(workdir),
            "Config Files (*.jsonc, *.json)",
        )
        if new_fn:
            save_json(Path(new_fn), jsonify(PyRandyOSApp.get_global_config()))

    @log_func_call
    def click_save_local_config(self):
        PyRandyOSApp.save_local_config()



# This file is auto-generated by pyrandyos/gui/icons/update_from_spec.py  # noqa: E501
# Do not edit.
_500px = 62062
accessible_icon = 62312
accusoft = 62313
acquisitions_incorporated = 63151
adn = 61808
adversal = 62314
affiliatetheme = 62315
airbnb = 63540
algolia = 62316
alipay = 63042
amazon = 62064
amazon_pay = 62508
amilia = 62317
android = 61819
angellist = 61961
angrycreative = 62318
angular = 62496
app_store = 62319
app_store_ios = 62320
apper = 62321
apple = 61817
apple_pay = 62485
artstation = 63354
asymmetrik = 62322
atlassian = 63355
audible = 62323
autoprefixer = 62492
avianex = 62324
aviato = 62497
aws = 62325
bandcamp = 62165
battle_net = 63541
behance = 61876
behance_square = 61877
bimobject = 62328
bitbucket = 61809
bitcoin = 62329
bity = 62330
black_tie = 62078
blackberry = 62331
blogger = 62332
blogger_b = 62333
bluetooth = 62099
bluetooth_b = 62100
bootstrap = 63542
btc = 61786
buffer = 63543
buromobelexperte = 62335
buy_n_large = 63654
buysellads = 61965
canadian_maple_leaf = 63365
cc_amazon_pay = 62509
cc_amex = 61939
cc_apple_pay = 62486
cc_diners_club = 62028
cc_discover = 61938
cc_jcb = 62027
cc_mastercard = 61937
cc_paypal = 61940
cc_stripe = 61941
cc_visa = 61936
centercode = 62336
centos = 63369
chrome = 62056
chromecast = 63544
cloudflare = 57469
cloudscale = 62339
cloudsmith = 62340
cloudversify = 62341
codepen = 61899
codiepie = 62084
confluence = 63373
connectdevelop = 61966
contao = 62061
cotton_bureau = 63646
cpanel = 62344
creative_commons = 62046
creative_commons_by = 62695
creative_commons_nc = 62696
creative_commons_nc_eu = 62697
creative_commons_nc_jp = 62698
creative_commons_nd = 62699
creative_commons_pd = 62700
creative_commons_pd_alt = 62701
creative_commons_remix = 62702
creative_commons_sa = 62703
creative_commons_sampling = 62704
creative_commons_sampling_plus = 62705
creative_commons_share = 62706
creative_commons_zero = 62707
critical_role = 63177
css3 = 61756
css3_alt = 62347
cuttlefish = 62348
d_and_d = 62349
d_and_d_beyond = 63178
dailymotion = 57426
dashcube = 61968
deezer = 57463
delicious = 61861
deploydog = 62350
deskpro = 62351
dev = 63180
deviantart = 61885
dhl = 63376
diaspora = 63377
digg = 61862
digital_ocean = 62353
discord = 62354
discourse = 62355
dochub = 62356
docker = 62357
draft2digital = 62358
dribbble = 61821
dribbble_square = 62359
dropbox = 61803
drupal = 61865
dyalog = 62361
earlybirds = 62362
ebay = 62708
edge = 62082
edge_legacy = 57464
elementor = 62512
ello = 62961
ember = 62499
empire = 61905
envira = 62105
erlang = 62365
ethereum = 62510
etsy = 62167
evernote = 63545
expeditedssl = 62014
facebook = 61594
facebook_f = 62366
facebook_messenger = 62367
facebook_square = 61570
fantasy_flight_games = 63196
fedex = 63383
fedora = 63384
figma = 63385
firefox = 62057
firefox_browser = 57351
first_order = 62128
first_order_alt = 62730
firstdraft = 62369
flickr = 61806
flipboard = 62541
fly = 62487
font_awesome = 62132
font_awesome_alt = 62300
font_awesome_flag = 62501
font_awesome_logo_full = 62694
fonticons = 62080
fonticons_fi = 62370
fort_awesome = 62086
fort_awesome_alt = 62371
forumbee = 61969
foursquare = 61824
free_code_camp = 62149
freebsd = 62372
fulcrum = 62731
galactic_republic = 62732
galactic_senate = 62733
get_pocket = 62053
gg = 62048
gg_circle = 62049
git = 61907
git_alt = 63553
git_square = 61906
github = 61595
github_alt = 61715
github_square = 61586
gitkraken = 62374
gitlab = 62102
gitter = 62502
glide = 62117
glide_g = 62118
gofore = 62375
goodreads = 62376
goodreads_g = 62377
google = 61856
google_drive = 62378
google_pay = 57465
google_play = 62379
google_plus = 62131
google_plus_g = 61653
google_plus_square = 61652
google_wallet = 61934
gratipay = 61828
grav = 62166
gripfire = 62380
grunt = 62381
guilded = 57470
gulp = 62382
hacker_news = 61908
hacker_news_square = 62383
hackerrank = 62967
hips = 62546
hire_a_helper = 62384
hive = 57471
hooli = 62503
hornbill = 62866
hotjar = 62385
houzz = 62076
html5 = 61755
hubspot = 62386
ideal = 57363
imdb = 62168
innosoft = 57472
instagram = 61805
instagram_square = 57429
instalod = 57473
intercom = 63407
internet_explorer = 62059
invision = 63408
ioxhost = 61960
itch_io = 63546
itunes = 62388
itunes_note = 62389
java = 62692
jedi_order = 62734
jenkins = 62390
jira = 63409
joget = 62391
joomla = 61866
js = 62392
js_square = 62393
jsfiddle = 61900
kaggle = 62970
keybase = 62709
keycdn = 62394
kickstarter = 62395
kickstarter_k = 62396
korvue = 62511
laravel = 62397
lastfm = 61954
lastfm_square = 61955
leanpub = 61970
less = 62493
line = 62400
linkedin = 61580
linkedin_in = 61665
linode = 62136
linux = 61820
lyft = 62403
magento = 62404
mailchimp = 62878
mandalorian = 62735
markdown = 62991
mastodon = 62710
maxcdn = 61750
mdb = 63690
medapps = 62406
medium = 62010
medium_m = 62407
medrt = 62408
meetup = 62176
megaport = 62883
mendeley = 63411
microblog = 57370
microsoft = 62410
mix = 62411
mixcloud = 62089
mixer = 57430
mizuni = 62412
modx = 62085
monero = 62416
napster = 62418
neos = 62994
nimblr = 62888
node = 62489
node_js = 62419
npm = 62420
ns8 = 62421
nutritionix = 62422
octopus_deploy = 57474
odnoklassniki = 62051
odnoklassniki_square = 62052
old_republic = 62736
opencart = 62013
openid = 61851
opera = 62058
optin_monster = 62012
orcid = 63698
osi = 62490
page4 = 62423
pagelines = 61836
palfed = 62424
patreon = 62425
paypal = 61933
penny_arcade = 63236
perbyte = 57475
periscope = 62426
phabricator = 62427
phoenix_framework = 62428
phoenix_squadron = 62737
php = 62551
pied_piper = 62126
pied_piper_alt = 61864
pied_piper_hat = 62693
pied_piper_pp = 61863
pied_piper_square = 57374
pinterest = 61650
pinterest_p = 62001
pinterest_square = 61651
playstation = 62431
product_hunt = 62088
pushed = 62433
python = 62434
qq = 61910
quinscape = 62553
quora = 62148
r_project = 62711
raspberry_pi = 63419
ravelry = 62169
react = 62491
reacteurope = 63325
readme = 62677
rebel = 61904
red_river = 62435
reddit = 61857
reddit_alien = 62081
reddit_square = 61858
redhat = 63420
renren = 61835
replyd = 62438
researchgate = 62712
resolving = 62439
rev = 62898
rocketchat = 62440
rockrms = 62441
rust = 57466
safari = 62055
salesforce = 63547
sass = 62494
schlix = 62442
scribd = 62090
searchengin = 62443
sellcast = 62170
sellsy = 61971
servicestack = 62444
shirtsinbulk = 61972
shopify = 57431
shopware = 62901
simplybuilt = 61973
sistrix = 62446
sith = 62738
sketch = 63430
skyatlas = 61974
skype = 61822
slack = 61848
slack_hash = 62447
slideshare = 61927
snapchat = 62123
snapchat_ghost = 62124
snapchat_square = 62125
soundcloud = 61886
sourcetree = 63443
speakap = 62451
speaker_deck = 63548
spotify = 61884
squarespace = 62910
stack_exchange = 61837
stack_overflow = 61804
stackpath = 63554
staylinked = 62453
steam = 61878
steam_square = 61879
steam_symbol = 62454
sticker_mule = 62455
strava = 62504
stripe = 62505
stripe_s = 62506
studiovinari = 62456
stumbleupon = 61860
stumbleupon_circle = 61859
superpowers = 62173
supple = 62457
suse = 63446
swift = 63713
symfony = 63549
teamspeak = 62713
telegram = 62150
telegram_plane = 62462
tencent_weibo = 61909
the_red_yeti = 63133
themeco = 62918
themeisle = 62130
think_peaks = 63281
tiktok = 57467
trade_federation = 62739
trello = 61825
tumblr = 61811
tumblr_square = 61812
twitch = 61928
twitter = 61593
twitter_square = 61569
typo3 = 62507
uber = 62466
ubuntu = 63455
uikit = 62467
umbraco = 63720
uncharted = 57476
uniregistry = 62468
unity = 57417
unsplash = 57468
untappd = 62469
ups = 63456
usb = 62087
usps = 63457
ussunnah = 62471
vaadin = 62472
viacoin = 62007
viadeo = 62121
viadeo_square = 62122
viber = 62473
vimeo = 62474
vimeo_square = 61844
vimeo_v = 62077
vine = 61898
vk = 61833
vnv = 62475
vuejs = 62495
watchman_monitoring = 57479
waze = 63551
weebly = 62924
weibo = 61834
weixin = 61911
whatsapp = 62002
whatsapp_square = 62476
whmcs = 62477
wikipedia_w = 62054
windows = 61818
wix = 62927
wizards_of_the_coast = 63280
wodu = 57480
wolf_pack_battalion = 62740
wordpress = 61850
wordpress_simple = 62481
wpbeginner = 62103
wpexplorer = 62174
wpforms = 62104
wpressr = 62436
xbox = 62482
xing = 61800
xing_square = 61801
y_combinator = 62011
yahoo = 61854
yammer = 63552
yandex = 62483
yandex_international = 62484
yarn = 63459
yelp = 61929
yoast = 62129
youtube = 61799
youtube_square = 62513
zhihu = 63039



from pathlib import Path
# from json import loads as jloads
from urllib.request import urlopen
from urllib.response import addinfourl

from ..logging import log_func_call
from .tqdm import optional_tqdm


@log_func_call
def get_github_download_url(base_url: str, commit: str, relpath: Path):
    return f'{base_url}/raw/{commit}/{relpath.as_posix()}'


@log_func_call
def download_file(url: str, dest: Path, use_tqdm: bool = True,
                  show_full_path: bool = False,
                  chunk_size: int = 8192):
    if dest.is_dir():
        dest = dest / Path(url).name

    with dest.open('wb') as f, urlopen(url) as response:
        response: addinfourl
        if response.status != 200:
            raise RuntimeError(f"Failed to download file from {url}: "
                               f"HTTP {response.status}")

        lbl = dest.as_posix() if show_full_path else dest.name
        with optional_tqdm(use_tqdm, desc=lbl, unit='B', unit_scale=True,
                           total=int(response.headers.get('Content-Length',
                                                          0))) as t:
            while True:
                chunk = response.read(chunk_size)
                if not chunk:
                    break
                f.write(chunk)
                t.update(len(chunk))

    return dest



from __future__ import annotations
from typing import TypeVar, Generic, TypeAlias

from ..qt import QCursor, QTimer
from .. import GuiQtView, GuiPresenter, QtWidgetType, GuiView
from ...logging import log_func_call

GuiWidgetPresType = TypeVar('GuiWidgetType', bound='GuiWidget')
GuiWidgetViewType = TypeVar('GuiWidgetViewType', bound='GuiWidgetView')
GuiWindowLikePresType = TypeVar('GuiWindowLikeType', bound='GuiWindowLike')
GuiWindowLikeViewType = TypeVar('GuiWindowLikeViewType',
                                bound='GuiWindowLikeView')

GuiWidgetParentType: TypeAlias = 'GuiWidget | GuiWidgetView | QtWidgetWrapper'
GuiWindowLikeParentType: TypeAlias = 'GuiWindowLike | GuiWindowLikeView'


class GuiWidget(GuiPresenter[GuiWidgetViewType], Generic[GuiWidgetViewType]):
    @log_func_call
    def __init__(self, gui_parent: GuiWidgetParentType | None = None,
                 *view_args, **view_kwargs):
        GuiPresenter.__init__(self, gui_parent, *view_args, **view_kwargs)

    @log_func_call
    def create_gui_view(self, *args, **kwargs) -> 'GuiWidgetViewType':
        return GuiWidgetView(self, *args, **kwargs)


class GuiSimpleWidgetPresenter(GuiWidget[None]):
    @log_func_call
    def create_gui_view(self, *args, **kwargs):
        pass


class GuiSimpleWidget(GuiSimpleWidgetPresenter, GuiView[None]):
    @log_func_call
    def __init__(self, gui_parent: GuiWidgetParentType | None = None,
                 *view_args, **view_kwargs):
        GuiSimpleWidgetPresenter.__init__(self, gui_parent, *view_args,
                                          **view_kwargs)
        GuiView.__init__(self, self)

    @log_func_call
    def create_gui_view(self, *args, **kwargs) -> None:
        raise NotImplementedError('Abstract method not implemented')


class GuiWidgetView(GuiQtView[GuiWidgetPresType, QtWidgetType],
                    Generic[GuiWidgetPresType, QtWidgetType]):
    @log_func_call
    def __init__(self, presenter: GuiWidgetPresType | None, *qtobj_args,
                 **qtobj_kwargs):
        GuiQtView.__init__(self, presenter, *qtobj_args, **qtobj_kwargs)


class QtWidgetWrapper(GuiSimpleWidgetPresenter,
                      GuiWidgetView[None, QtWidgetType],
                      Generic[QtWidgetType]):
    @log_func_call
    def __init__(self, gui_parent: GuiWidgetParentType | None = None,
                 *qtobj_args, **qtobj_kwargs):
        GuiSimpleWidgetPresenter.__init__(self, gui_parent)
        GuiWidgetView.__init__(self, self, *qtobj_args, **qtobj_kwargs)


class GuiWindowLike(GuiWidget[GuiWindowLikeViewType],
                    Generic[GuiWindowLikeViewType]):
    @log_func_call
    def __init__(self, basetitle: str,
                 gui_parent: GuiWidgetParentType | None = None,
                 *view_args, **view_kwargs):
        super().__init__(gui_parent, basetitle, *view_args, **view_kwargs)
        # self.gui_view: GuiWindowLikeView

    @log_func_call
    def show(self):
        self.gui_view.show()

    @log_func_call
    def create_gui_view(self, basetitle: str, *args,
                        **kwargs) -> GuiWindowLikeViewType:
        raise NotImplementedError('Abstract method not implemented')

    @log_func_call
    def post_show_init(self):
        "Any initialization that should be executed after the window is shown"
        pass


class GuiWindowLikeView(GuiWidgetView[GuiWindowLikePresType, QtWidgetType],
                        Generic[GuiWindowLikePresType, QtWidgetType]):
    @log_func_call
    def __init__(self, basetitle: str, presenter: GuiWindowLikePresType = None,
                 *qtobj_args, **qtobj_kwargs):
        super().__init__(presenter, *qtobj_args, **qtobj_kwargs)
        self.basetitle = basetitle
        self.set_title()
        QTimer.singleShot(0, presenter.post_show_init)

    @log_func_call
    def show(self):
        self.qtobj.show()

    @property
    def hwnd(self):
        return self.qtobj.winId()

    @log_func_call
    def bring_to_front(self):
        qtobj = self.qtobj
        qtobj.raise_()
        qtobj.activateWindow()
        qtobj.showNormal()

    @log_func_call
    def center_window_in_current_screen(self):
        qtobj = self.qtobj
        qtapp = self.qt_app
        centerPoint = qtapp.screenAt(QCursor.pos()).geometry().center()
        geo = qtobj.geometry()
        geo.moveCenter(centerPoint)
        qtobj.setGeometry(geo)

    @log_func_call
    def set_title(self, subtitle: str = None):
        title = self.basetitle
        if subtitle:
            title += ' - ' + subtitle

        self.qtobj.setWindowTitle(title)

    @log_func_call
    def get_dpi(self):
        return self.qtobj.devicePixelRatio()



"""
This module ensures that the Qt resource file (resources.qrc) is compiled to a
Python module (qt_resources.py) at runtime if it is missing or out of date.
It should be imported before any Qt widgets are created.
"""
from subprocess import run, CalledProcessError
from pathlib import Path

from ..logging import log_func_call
from ..config.keys import QRC_PYFILE_KEY, QRC_FILE_KEY
from ..utils.constants import IS_WIN32
from ..utils.system import import_python_file
from ..app import PyRandyOSApp

HERE = Path(__file__).parent
PY_FILENAME = "qt_resources.py"
QT_RESOURCES_MODULE = f"{__name__}.resources"


@log_func_call
def compile_qrc(qrcfile: Path = None):
    """Compile the Qt resource file if needed."""
    if qrcfile:
        PyRandyOSApp.set(QRC_FILE_KEY, qrcfile)
    else:
        qrcfile: Path = qrcfile or PyRandyOSApp.get(QRC_FILE_KEY, None)

    if not qrcfile:
        return
    if qrcfile and not qrcfile.exists():
        raise FileNotFoundError(f"Resource file not found: {qrcfile}")

    tmpdir = PyRandyOSApp.mkdir_temp()
    py_file = tmpdir/PY_FILENAME
    PyRandyOSApp.set(QRC_PYFILE_KEY, py_file)

    # Only recompile if .py is missing or older than .qrc
    if py_file.exists() and py_file.stat().st_mtime > qrcfile.stat().st_mtime:
        return  # Up to date

    # Try to find pyside2-rcc or pyrcc5
    rcc_cmds = [
        ["pyside2-rcc", str(qrcfile)],
        ["pyrcc5", str(qrcfile)],
    ]
    for cmd in rcc_cmds:
        try:
            # On Windows, need to add shell=True for .bat/.cmd
            result = run(cmd, capture_output=True, check=True, shell=IS_WIN32)
            if result.returncode == 0:
                py_file.write_bytes(result.stdout)
                return
        except (FileNotFoundError, CalledProcessError):
            continue
    raise RuntimeError(
        "Could not compile Qt resources. "
        "Please ensure 'pyside2-rcc' or 'pyrcc5' is installed and on PATH."
    )


@log_func_call
def import_qrc():
    # Import the generated resource module so resources are registered
    qrcfile: Path = PyRandyOSApp.get(QRC_FILE_KEY, None)
    if not qrcfile:
        return
    py_file: Path = PyRandyOSApp.get(QRC_PYFILE_KEY, None)
    if not py_file or not py_file.exists():
        compile_qrc()

    try:
        mod = import_python_file(py_file, QT_RESOURCES_MODULE)
    except ImportError:
        mod = None

    if not mod:
        raise ImportError("Could not import compiled Qt resource module: "
                          f"{py_file}")



# This file is auto-generated by pyrandyos/gui/icons/update_from_spec.py  # noqa: E501
# Do not edit.
from ....iconfont import IconFont


class FluentUI_Filled(IconFont):
    _SPECNAME = 'FluentUI.Filled'



import sys
from pathlib import Path
from re import compile, match, MULTILINE

from pyrandyos.utils.json import parse_jsonc


DECORATOR = '@log_func_call'

DECORATOR_RE = compile(r'^[ \t]*@[_]?log_func_call(\s*\(.*\))?\s*$',
                       flags=MULTILINE)
CONFIG_FILENAME = ".log_func_call_exclude.jsonc"


def load_exclude_config(config_path):
    config_path = Path(config_path)
    if not config_path.exists():
        return {
            "exclude_files": [],
            "exclude_classes": [],
            "exclude_methods": {},
            "exclude_dirs": [],
            "base_dir": None
        }
    with open(config_path, "r", encoding="utf-8") as f:
        return parse_jsonc(f.read())


def should_skip_dir(path: Path, exclude_dirs) -> bool:
    return any(part in exclude_dirs for part in path.parts)


def is_property_decorator(line):
    stripped = line.strip()
    return (
        stripped == '@property'
        or stripped.endswith('.setter')
        or stripped.endswith('.deleter')
    )


def file_needs_decorator(filepath: Path, exclude_file_paths, exclude_classes,
                         exclude_methods, exclude_dirs, should_skip_dir):
    content = filepath.read_text(encoding='utf-8')
    lines = content.splitlines()
    new_lines = []
    i = 0
    changed = False
    scope_stack = []
    while i < len(lines):
        line = lines[i]
        indent = len(line) - len(line.lstrip())

        # Pop scopes if indentation decreases (ignore blank lines)
        while scope_stack and line.strip() and indent <= scope_stack[-1][2]:
            scope_stack.pop()

        # Enter TYPE_CHECKING block
        type_checking_match = match(r'^([ \t]*)if TYPE_CHECKING:', line)
        if type_checking_match:
            scope_stack.append(('type_checking', None, indent))

        # Enter class
        class_match = match(r'^([ \t]*)class ([a-zA-Z0-9_]+)', line)
        if class_match:
            class_name = class_match.group(2)
            scope_stack.append(('class', class_name, indent))

        # Check if inside TYPE_CHECKING
        inside_type_checking = any(
            s[0] == 'type_checking' for s in scope_stack
        )

        # Check if inside excluded class
        inside_excluded_class = False
        current_class = None
        for s in reversed(scope_stack):
            if s[0] == 'class':
                current_class = s[1]
                if current_class in exclude_classes:
                    inside_excluded_class = True
                break

        # Exclude specific methods
        func_match = match(r'^[ \t]*def ([a-zA-Z0-9_]+)', line)
        if func_match:
            method_name = func_match.group(1)
            if (current_class and current_class in exclude_methods and
                    method_name in exclude_methods[current_class]):
                new_lines.append(line)
                i += 1
                new_lines.append("")
                continue

        # Exclude functions in excluded classes or TYPE_CHECKING
        if match(r'^[ \t]*def [a-zA-Z0-9_]+', line):
            if inside_excluded_class or inside_type_checking:
                new_lines.append(line)
                i += 1
                new_lines.append("")
                continue

            # Scan upwards for decorators
            j = len(new_lines) - 1
            has_decorator = False
            is_property = False
            while j >= 0 and (new_lines[j].strip() == ''
                              or new_lines[j].lstrip().startswith('@')):
                if DECORATOR_RE.match(new_lines[j]):
                    has_decorator = True
                if is_property_decorator(new_lines[j]):
                    is_property = True
                j -= 1

            if not has_decorator and not is_property:
                indent_str = match(r'^([ \t]*)', line).group(1)
                new_lines.append(f'{indent_str}{DECORATOR}')
                changed = True

        new_lines.append(line)
        i += 1

    if changed:
        filepath.write_text('\n'.join(new_lines) + '\n', encoding='utf-8')
        print(f'Updated: {filepath}')


def main(root_dirs, config_path=None):
    script_path = Path(__file__).resolve()

    if config_path:
        config = load_exclude_config(config_path)
        base_path = config.get("base_dir")
        if base_path:
            base_path = Path(base_path).resolve()
            root_dirs = [str(base_path)]  # Use base_dir as only search root
        else:
            base_path = Path.cwd().resolve()
        configs = [config]
    else:
        configs = []
        base_path = Path.cwd().resolve()
        for root_dir in root_dirs:
            config = load_exclude_config(Path(root_dir) / CONFIG_FILENAME)
            configs.append(config)

    exclude_files = set()
    exclude_classes = set()
    exclude_methods = dict()
    exclude_dirs = set()
    exclude_functions_by_file = dict()

    for config in configs:
        exclude_files.update(config.get("exclude_files", []))
        exclude_classes.update(config.get("exclude_classes", []))
        exclude_dirs.update(config.get("exclude_dirs", []))
        # exclude_functions is now a dict: {filename: [func1, func2, ...]}
        ef = config.get("exclude_functions", {})
        for fname, funclist in ef.items():
            if fname not in exclude_functions_by_file:
                exclude_functions_by_file[fname] = set()
            exclude_functions_by_file[fname].update(funclist)

        for cls, methods in config.get("exclude_methods", {}).items():
            if cls not in exclude_methods:
                exclude_methods[cls] = set()
            exclude_methods[cls].update(methods)

    exclude_file_paths = set()
    for p in exclude_files:
        path_obj = Path(p)
        if not path_obj.is_absolute():
            path_obj = (base_path / path_obj).resolve()
        exclude_file_paths.add(str(path_obj))

    def should_skip_dir(path: Path) -> bool:
        return any(part in exclude_dirs for part in path.parts)

    def file_needs_decorator(filepath: Path):
        content = filepath.read_text(encoding='utf-8')
        lines = content.splitlines()
        new_lines = []
        i = 0
        changed = False
        scope_stack = []
        # Compute relative path from base_dir for this file
        rel_path = str(filepath.resolve().relative_to(base_path))
        rel_path = rel_path.replace('\\', '/').replace('\\', '/')
        exclude_funcs_for_file = exclude_functions_by_file.get(rel_path, set())

        while i < len(lines):
            line = lines[i]
            indent = len(line) - len(line.lstrip())

            # Pop scopes if indentation decreases (ignore blank lines)
            while (scope_stack and line.strip()
                   and indent <= scope_stack[-1][2]):
                scope_stack.pop()

            # Enter TYPE_CHECKING block
            type_checking_match = match(r'^([ \t]*)if TYPE_CHECKING:', line)
            if type_checking_match:
                scope_stack.append(('type_checking', None, indent))

            # Enter class
            class_match = match(r'^([ \t]*)class ([a-zA-Z0-9_]+)', line)
            if class_match:
                class_name = class_match.group(2)
                scope_stack.append(('class', class_name, indent))

            # Check if inside TYPE_CHECKING
            inside_type_checking = any(
                s[0] == 'type_checking' for s in scope_stack
            )

            # Check if inside excluded class
            inside_excluded_class = False
            current_class = None
            for s in reversed(scope_stack):
                if s[0] == 'class':
                    current_class = s[1]
                    if current_class in exclude_classes:
                        inside_excluded_class = True
                    break

            # Exclude specific methods or module-level functions
            func_match = match(r'^[ \t]*def ([a-zA-Z0-9_]+)', line)
            if func_match:
                method_name = func_match.group(1)
                if current_class:
                    if (current_class in exclude_methods and
                            method_name in exclude_methods[current_class]):
                        new_lines.append(line)
                        i += 1
                        new_lines.append("")
                        continue
                else:
                    # module-level function
                    if method_name in exclude_funcs_for_file:
                        new_lines.append(line)
                        i += 1
                        new_lines.append("")
                        continue

            # Exclude functions in excluded classes or TYPE_CHECKING
            if match(r'^[ \t]*def [a-zA-Z0-9_]+', line):
                if inside_excluded_class or inside_type_checking:
                    new_lines.append(line)
                    i += 1
                    new_lines.append("")
                    continue

                # Scan upwards for decorators
                j = len(new_lines) - 1
                has_decorator = False
                is_property = False
                while j >= 0 and (new_lines[j].strip() == ''
                                  or new_lines[j].lstrip().startswith('@')):
                    if DECORATOR_RE.match(new_lines[j]):
                        has_decorator = True
                    if is_property_decorator(new_lines[j]):
                        is_property = True
                    j -= 1
                if not has_decorator and not is_property:
                    indent_str = match(r'^([ \t]*)', line).group(1)
                    new_lines.append(f'{indent_str}{DECORATOR}')
                    changed = True
            new_lines.append(line)
            i += 1
        if changed:
            filepath.write_text('\n'.join(new_lines) + '\n', encoding='utf-8')
            print(f'Updated: {filepath}')

    for root_dir in root_dirs:
        for pyfile in Path(root_dir).rglob('*.py'):
            resolved_pyfile = pyfile.resolve()

            if resolved_pyfile == script_path:
                continue

            if (
                pyfile.name in exclude_files
                or str(resolved_pyfile) in exclude_file_paths
            ):
                continue

            if should_skip_dir(pyfile):
                continue

            file_needs_decorator(pyfile)


if __name__ == '__main__':
    if len(sys.argv) > 1:
        config_path = sys.argv[1]
        main(['.'], config_path=config_path)
    else:
        main(['.'])



from typing import TYPE_CHECKING

from ....logging import log_func_call
from ...qt import (
    QVBoxLayout, QDialogButtonBox, QTextEdit, QTableWidget,
    QTableWidgetItem, QHeaderView, Qt, QFont,
)
from .. import GuiDialogView
from ...callback import qt_callback

if TYPE_CHECKING:
    from .pres import LogDialog

# Column definitions: (label, autofit)
LOG_COLS = [
    ("Timestamp", True),
    ("Level", True),
    ("File", True),
    ("Function", True),
    ("Message", False),
]

UserRole = Qt.UserRole

MONO_FONT = QFont("Consolas, Monaco, monospace")
MONO_FONT.setStyleHint(QFont.Monospace)
TABLE_TPAD = 2  # one sided
TABLE_LPAD = 4  # one sided
TABLE_VPAD = TABLE_TPAD*2  # two sided
TABLE_HPAD = TABLE_LPAD*2  # two sided
TABLE_CELL_STYLE = f"QTableWidget::item {{ padding: {TABLE_TPAD}px {TABLE_LPAD}px; }}"  # noqa: E501


class LogDialogView(GuiDialogView['LogDialog']):
    @log_func_call
    def __init__(self, basetitle: str, presenter: 'LogDialog' = None,
                 *qtobj_args, **qtobj_kwargs):
        GuiDialogView.__init__(self, basetitle, presenter, *qtobj_args,
                               **qtobj_kwargs)
        qtobj = self.qtobj
        qtobj.resize(1500, 600)
        self.layout = QVBoxLayout(qtobj)
        self.create_log_widget()

    @log_func_call
    def create_log_widget(self):
        __traceback_hide_locals__ = 'content'  # noqa: F841
        qtobj = self.qtobj
        layout = self.layout
        pres = self.gui_pres

        content = pres.parse_log_file()
        self.log_table = None
        if isinstance(content, str):
            # Create scrollable text widget for log content
            log_text = QTextEdit()
            log_text.setReadOnly(True)
            log_text.setLineWrapMode(QTextEdit.NoWrap)
            log_text.setPlainText(content)
            layout.addWidget(log_text)
            self.log_text = log_text

            # Scroll to bottom to show most recent entries
            cursor = log_text.textCursor()
            cursor.movePosition(cursor.End)
            log_text.setTextCursor(cursor)

        else:
            log_table = QTableWidget()
            layout.addWidget(log_table)
            self.log_table = log_table
            self.setup_log_table()
            # need to wait until after __init__ to update row heights
            self.populate_log_table(content, False)

        dlgbuttons = QDialogButtonBox(QDialogButtonBox.Ok, qtobj)
        dlgbuttons.accepted.connect(qtobj.accept)
        layout.addWidget(dlgbuttons)
        self.dlgbuttons = dlgbuttons

    @log_func_call
    def setup_log_table(self):
        pres = self.gui_pres
        table = self.log_table

        header = table.horizontalHeader()

        table.setColumnCount(len(LOG_COLS))
        for i, (label, autofit) in enumerate(LOG_COLS):
            item = QTableWidgetItem(label)
            table.setHorizontalHeaderItem(i, item)
            # if autofit:
            #     # Auto-resize to fit content for fixed columns
            #     header.setSectionResizeMode(i, QHeaderView.ResizeToContents)
            # else:
            #     # Start with Stretch, will switch to Interactive
            #     header.setSectionResizeMode(i, QHeaderView.Stretch)

        header.setSectionResizeMode(QHeaderView.ResizeToContents)
        header.setStretchLastSection(True)
        # Add padding that works with stretch columns
        table.setStyleSheet(TABLE_CELL_STYLE)
        table.setSortingEnabled(False)
        table.setAlternatingRowColors(True)
        table.setSelectionBehavior(QTableWidget.SelectRows)
        table.setEditTriggers(QTableWidget.NoEditTriggers)  # Make readonly
        table.setWordWrap(True)
        table.setVerticalScrollMode(QTableWidget.ScrollPerPixel)
        table.setHorizontalScrollMode(QTableWidget.ScrollPerPixel)

        vheader = table.verticalHeader()
        # vheader.setDefaultSectionSize(20)
        # vheader.setMinimumSectionSize(20)
        # vheader.resizeSections(QHeaderView.ResizeToContents)
        vheader.setSectionResizeMode(QHeaderView.Fixed)
        vheader.sectionClicked.connect(qt_callback(pres.row_header_clicked))

        # don't do this yet because sorting later will trigger it
        # self.connect_column_resize_signal()

    @log_func_call
    def connect_column_resize_signal(self):
        header = self.log_table.horizontalHeader()
        cb = qt_callback(self.gui_pres.update_row_heights)
        header.sectionResized.connect(cb)

    @log_func_call
    def disconnect_column_resize_signal(self, raise_exc: bool = True):
        header = self.log_table.horizontalHeader()
        try:
            header.sectionResized.disconnect(self.gui_pres.update_row_heights)
        except RuntimeError:
            if raise_exc:
                raise

    @log_func_call
    def populate_log_table(self, content: tuple,
                           update_row_heights: bool = True):
        __traceback_hide_locals__ = ('content', 'col')  # noqa: F841
        # timestamps, levels, files, funcs, msgs = content
        table = self.log_table

        # disable resize signals while populating since sort will trigger
        self.disconnect_column_resize_signal(False)

        rowcount = len(content[0])
        table.setRowCount(rowcount)
        for row in range(rowcount):
            # Create regular table items and store original text as user data
            for i, col in enumerate(content):
                x = col[row]
                item = QTableWidgetItem(x)
                item.setData(UserRole, x)
                # Apply monospace font to the message column (last column)
                if i == len(LOG_COLS) - 1:  # Message column
                    item.setFont(MONO_FONT)

                table.setItem(row, i, item)

        table.sortItems(0, Qt.DescendingOrder)

        # After populating data, ensure proper column sizing
        # header = table.horizontalHeader()
        # for i, (_, autofit) in enumerate(LOG_COLS):
        #     if not autofit:
        #         # For non-autofit columns (Message), ensure Interactive mode
        #         # and set a reasonable initial width if needed
        #         header.setSectionResizeMode(i, QHeaderView.Interactive)
        #         # if header.sectionSize(i) < 200:  # Min width for msg col
        #         #     header.resizeSection(i, 400)  # Set reasonable default

        # Update row heights after populating and setting up columns
        if update_row_heights:
            self.gui_pres.update_row_heights()

        # reconnect signal
        self.connect_column_resize_signal()



import sys


def is_debug_enabled():
    # see https://stackoverflow.com/a/77627047/13230486
    try:
        if sys.gettrace() is not None:
            return True
    except AttributeError:
        pass

    try:
        if sys.monitoring.get_tool(sys.monitoring.DEBUGGER_ID) is not None:
            return True
    except AttributeError:
        pass

    return False


def hide_splash():
    from pyrandyos.gui.gui_app import get_gui_app
    gui_app = get_gui_app()
    if gui_app and gui_app.splash:
        splash = gui_app.splash.gui_view.qtobj
        if splash.isVisible():
            splash.hide()
            if is_debug_enabled():
                print("Splash screen hidden, debugger is enabled")
            else:
                print("Splash screen hidden")
        else:
            print("Splash screen is already hidden")
    else:
        print("No splash screen to hide")



# see: https://stackoverflow.com/a/5191224/13230486

from collections.abc import Callable


class ClassPropertyDescriptor:
    def __init__(self, fget: Callable, fset: Callable = None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")

        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func: Callable):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


def classproperty(func: Callable):
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)



"""
This module is intended for internal use only.  No user documentation is
provided at this time.  Use at your own discretion.
"""

# code adapted from https://github.com/Drekin/win-unicode-console/
from logging import Logger
from pathlib import Path
from ctypes import c_ulong, byref, create_string_buffer

from ...logging import log_func_call
from ..constants import IS_WIN32
from ..constants.windows import (
    ErrCodeLookup, DllErrMsgDict, DriveList, ERROR_INVALID_HANDLE,
    PROCESS_SYSTEM_DPI_AWARE, MAX_PREFERRED_LENGTH,
)
from .ctypes import IS_WIN_CTYPES

if IS_WIN_CTYPES:
    from .ctypes import (
        WriteConsoleW, GetConsoleMode, get_last_error, WNetOpenEnum,
        WNetCloseEnum, WNetEnumResource, WinError, SetProcessDpiAwareness,
        RESOURCETYPE_DISK, RESOURCE_CONNECTED,
    )


@log_func_call
def build_error_message(codelookup: ErrCodeLookup, dllerrs: DllErrMsgDict,
                        func: str, code: int,):
    errname = codelookup.get(code)
    funcerrs = dllerrs.get(func)
    errmsg = funcerrs.get(errname) if funcerrs else None

    msg = f'{func} error: {errname or "[Unknown]"}({hex(code)})'
    if errmsg:
        msg += f': {errmsg}'

    return msg


@log_func_call
def win_write(handle: int, data: bytes):
    bytes_to_be_written = len(data)
    buffer = create_string_buffer(data)
    code_units_to_be_written = bytes_to_be_written//2
    code_units_written = c_ulong()

    if not WriteConsoleW(handle, buffer, code_units_to_be_written,
                         byref(code_units_written), None):
        exc = WinError(get_last_error())
        raise exc  # use traceback from here

    return 2 * code_units_written.value  # bytes written


@log_func_call
def is_win_console(handle: int):
    mode = c_ulong()
    retval = GetConsoleMode(handle, byref(mode))
    return bool(retval) or get_last_error() != ERROR_INVALID_HANDLE


@log_func_call
def set_high_dpi_support(value: int = PROCESS_SYSTEM_DPI_AWARE,
                         log: Logger = None):
    if IS_WIN_CTYPES:
        if log:
            log.debug(f'setting process to DPI aware ({value})')

        SetProcessDpiAwareness(value)


@log_func_call
def get_mapped_drives():
    """
    Gets all currently mapped drives in Windows.

    Returns:
        DriveList: list of tuples of mapped drives.
            The two Paths in the tuple are (local, remote), where the local
            is a drive letter and remote is the UNC of the network share.
    """
    if not IS_WIN32:
        return list()
    handle = WNetOpenEnum(RESOURCE_CONNECTED, RESOURCETYPE_DISK, 0, None)
    drives: DriveList = list()
    try:
        items = True
        while items:
            items = WNetEnumResource(handle, MAX_PREFERRED_LENGTH)
            drives += [(Path(it.lpLocalName), Path(it.lpRemoteName))
                       for it in items]
            # for it in items:
            #     print(it.dwScope)
            #     print(it.dwType)
            #     print(it.dwDisplayType)
            #     print(RESOURCEDISPLAYTYPES[it.dwDisplayType])
            #     print(it.dwUsage)
            #     print(it.lpLocalName)
            #     print(it.lpRemoteName)
            #     print(it.lpComment)
            #     print(it.lpProvider)

        return drives
    finally:
        WNetCloseEnum(handle)



from collections.abc import Callable

from ..logging import log_func_call
from ..utils.stack import get_stack_frame, get_framesummary_for_frame


class QtCallable:
    def __init__(self, func: Callable, stacklevel: int = 1):
        self.func = func
        # stacklevel arg should account for this function at a minimum, hence 1
        # but then we need to add 1 for get_stack_frame.  This should end up
        # returning the frame that called the QtCallable ctor
        fs = get_framesummary_for_frame(get_stack_frame(stacklevel + 1))
        self.caller = fs
        self.caller_line = fs.line

    def __call__(self, *args, **kwargs):
        caller = self.caller  # noqa: F841
        caller_line = self.caller_line  # noqa: F841
        return log_func_call(self.func)(*args, **kwargs)


def qt_callback(f: Callable):
    # increment the stacklevel from default by 1 to account for this frame
    return QtCallable(f, 2)



from __future__ import annotations
from typing import TypeVar, Generic

from ..logging import log_func_call
from .widgets import GuiWindowLikeView, GuiWindowLike, GuiWindowLikeParentType
from .widgets.viewbase import GuiViewBaseWidget
from .qt import QMainWindow

GuiWindowPresType = TypeVar('GuiWindowPresType', bound='GuiWindow')
GuiWindowViewType = TypeVar('GuiWindowViewType', bound='GuiWindowView')
GuiViewBaseWidgetType = TypeVar('GuiViewBaseWidgetType',
                                bound=GuiViewBaseWidget)


class GuiWindow(GuiWindowLike[GuiWindowViewType], Generic[GuiWindowViewType]):
    @log_func_call
    def __init__(self, basetitle: str,
                 gui_parent: GuiWindowLikeParentType | None = None,
                 *view_args, **view_kwargs):
        super().__init__(basetitle, gui_parent, *view_args, **view_kwargs)


class GuiWindowView(GuiWindowLikeView[GuiWindowPresType, QMainWindow],
                    Generic[GuiWindowPresType, GuiViewBaseWidgetType]):
    @log_func_call
    def __init__(self, basetitle: str, presenter: GuiWindowPresType | None,
                 *qtobj_args, **qtobj_kwargs):
        GuiWindowLikeView.__init__(self, basetitle, presenter, *qtobj_args,
                                   **qtobj_kwargs)
        qtobj = self.qtobj
        basewidget = self.create_basewidget()
        self._basewidget = basewidget
        qtobj.setCentralWidget(basewidget.qtobj)

    @log_func_call
    def create_qtobj(self, *args, **kwargs):
        return QMainWindow(*args, **kwargs)

    @log_func_call
    def create_basewidget(self) -> GuiViewBaseWidgetType:
        raise NotImplementedError('Abstract method not implemented')

    @property
    def basewidget(self) -> GuiViewBaseWidgetType:
        return self._basewidget

    # @property
    # def qtbasewidget(self):
    #     return self.basewidget.qtobj



from pathlib import Path

from .fontspec import (
    IconCharMapFileSpec, IconFontGitCommit, IconFontSpec, IconTtfFileSpec,
)

QtAwesome = IconFontGitCommit(
    "https://github.com/spyder-ide/qtawesome",
    "b79946412c7ba58c739082634c8d67f99773cc4b",
    Path("LICENSE.txt")
)

Codicons_repo = IconFontGitCommit(
    "https://github.com/microsoft/vscode-codicons",
    "3f96711c6f617f419d1a84a4d86d92021fc7bda6",
    (Path("LICENSE"), Path("LICENSE-CODE"))
)

VSCode = IconFontGitCommit(
    "https://github.com/microsoft/vscode",
    "dfe124cea49046c8ad5d7902b3f05a60d27bc70b",
    Path("LICENSE.txt")
)

FluentUI_repo = IconFontGitCommit(
    "https://github.com/microsoft/fluentui-system-icons",
    "5baf88376599abf9914416c95d050c24c55f7000",
    Path("LICENSE")
)

qta_font_dir = Path("qtawesome/fonts")
fluent_font_dir = Path("fonts")


THIRDPARTY_FONTSPEC = {
    "Codicons": IconFontSpec(
        IconTtfFileSpec(VSCode, Path("src/vs/base/browser/ui/codicons/codicon/codicon.ttf"), "5608b14a69ae55c99a67ef20249a15c7"),  # noqa: E501
        IconCharMapFileSpec(Codicons_repo, Path("src/template/mapping.json"), 10, "8df7fbe29b47eff0917d6ecef17c5855"),  # noqa: E501
    ),
    "ElusiveIcons": IconFontSpec(
        IconTtfFileSpec(QtAwesome, qta_font_dir/"elusiveicons-webfont-2.0.ttf", "207966b04c032d5b873fd595a211582e"),  # noqa: E501
        IconCharMapFileSpec(QtAwesome, qta_font_dir/"elusiveicons-webfont-charmap-2.0.json", 16, "c3b93ddda4215b385290f85be52619ff"),  # noqa: E501
    ),
    "Fa5.Brands": IconFontSpec(
        IconTtfFileSpec(QtAwesome, qta_font_dir/"fontawesome5-brands-webfont-5.15.4.ttf", "513aa607d398efaccc559916c3431403"),  # noqa: E501
        IconCharMapFileSpec(QtAwesome, qta_font_dir/"fontawesome5-brands-webfont-charmap-5.15.4.json", 16, "5c56bff92bfdc668239f5acb9f3e9fdf")  # noqa: E501
    ),
    "Fa5": IconFontSpec(
        IconTtfFileSpec(QtAwesome, qta_font_dir/"fontawesome5-regular-webfont-5.15.4.ttf", "dc47e4089f5bcb25f241bdeb2de0fb58"),  # noqa: E501
        IconCharMapFileSpec(QtAwesome, qta_font_dir/"fontawesome5-regular-webfont-charmap-5.15.4.json", 16, "0c91f5f5feeaba3b1dfc825565258cee"),  # noqa: E501
    ),
    "Fa5.Solid": IconFontSpec(
        IconTtfFileSpec(QtAwesome, qta_font_dir/"fontawesome5-solid-webfont-5.15.4.ttf", "5de19800fc9ae73438c2e5c61d041b48"),  # noqa: E501
        IconCharMapFileSpec(QtAwesome, qta_font_dir/"fontawesome5-solid-webfont-charmap-5.15.4.json", 16, "e71dcf48ec1e03a2034ff38d94bc5aca"),  # noqa: E501
        solid=True,
    ),
    "Fa6.Brands": IconFontSpec(
        IconTtfFileSpec(QtAwesome, qta_font_dir/"fontawesome6-brands-webfont-6.7.2.ttf", "15d54d142da2f2d6f2e90ed1d55121af"),  # noqa: E501
        IconCharMapFileSpec(QtAwesome, qta_font_dir/"fontawesome6-brands-webfont-charmap-6.7.2.json", 16, "071f8106a165c7f69310d09d6802cd47"),  # noqa: E501
    ),
    "Fa6": IconFontSpec(
        IconTtfFileSpec(QtAwesome, qta_font_dir/"fontawesome6-regular-webfont-6.7.2.ttf", "2b9e6cb53822f6a9b42f15229a36811a"),  # noqa: E501
        IconCharMapFileSpec(QtAwesome, qta_font_dir/"fontawesome6-regular-webfont-charmap-6.7.2.json", 16, "2c6bc833abc45cdda3fe5c2249104771"),  # noqa: E501
    ),
    "Fa6.Solid": IconFontSpec(
        IconTtfFileSpec(QtAwesome, qta_font_dir/"fontawesome6-solid-webfont-6.7.2.ttf", "07312c769a05b2c17133da1a09db4ccf"),  # noqa: E501
        IconCharMapFileSpec(QtAwesome, qta_font_dir/"fontawesome6-solid-webfont-charmap-6.7.2.json", 16, "0677f8ff2f9ed4ef28ad421129508309"),  # noqa: E501
        solid=True,
    ),
    "FluentUI.Filled": IconFontSpec(
        IconTtfFileSpec(FluentUI_repo, fluent_font_dir/"FluentSystemIcons-Filled.ttf", "e3282e84d8d46e5024d6547c11b7900b"),  # noqa: E501
        IconCharMapFileSpec(FluentUI_repo, fluent_font_dir/"FluentSystemIcons-Filled.json", 10, "82a97b64b6e2af20f28155793dcb5a1c"),  # noqa: E501
    ),
    "FluentUI": IconFontSpec(
        IconTtfFileSpec(FluentUI_repo, fluent_font_dir/"FluentSystemIcons-Regular.ttf", "f3d8cc1d4f6435fe29d8c3b619047dec"),  # noqa: E501
        IconCharMapFileSpec(FluentUI_repo, fluent_font_dir/"FluentSystemIcons-Regular.json", 10, "0020dbda23979b4d876252f938c01e50"),  # noqa: E501
    ),
    "FluentUI.Light": IconFontSpec(
        IconTtfFileSpec(FluentUI_repo, fluent_font_dir/"FluentSystemIcons-Light.ttf", "d5c442567a0b0bdf068a96ebaba04748"),  # noqa: E501
        IconCharMapFileSpec(FluentUI_repo, fluent_font_dir/"FluentSystemIcons-Light.json", 10, "e9839f6ed1cf395e42f7725a0b4c1aab"),  # noqa: E501
    ),
    "FluentUI.Resize": IconFontSpec(
        IconTtfFileSpec(FluentUI_repo, fluent_font_dir/"FluentSystemIcons-Resizable.ttf", "b0897636fcc80e20b5dc5fd2283b268d"),  # noqa: E501
        IconCharMapFileSpec(FluentUI_repo, fluent_font_dir/"FluentSystemIcons-Resizable.json", 10, "6e53427d2b5efa09c0c1379a0f42b0a9"),  # noqa: E501
    ),
    "Material5": IconFontSpec(
        IconTtfFileSpec(QtAwesome, qta_font_dir/"materialdesignicons5-webfont-5.9.55.ttf", "b7d40e7ef80c1d4af6d94902af66e524"),  # noqa: E501
        IconCharMapFileSpec(QtAwesome, qta_font_dir/"materialdesignicons5-webfont-charmap-5.9.55.json", 16, "0e688abb70cf39c1661f4a25bd21d46c"),  # noqa: E501
    ),
    "Material6": IconFontSpec(
        IconTtfFileSpec(QtAwesome, qta_font_dir/"materialdesignicons6-webfont-6.9.96.ttf", "ecaabfbb23fdac4ddbaf897b97257a92"),  # noqa: E501
        IconCharMapFileSpec(QtAwesome, qta_font_dir/"materialdesignicons6-webfont-charmap-6.9.96.json", 16, "c458eb03521a472a03ffdfe942cfb0c0"),  # noqa: E501
    ),
    "Phosphor": IconFontSpec(
        IconTtfFileSpec(QtAwesome, qta_font_dir/"phosphor-1.3.0.ttf", "5b8dc57388b2d86243566b996cc3a789"),  # noqa: E501
        IconCharMapFileSpec(QtAwesome, qta_font_dir/"phosphor-charmap-1.3.0.json", 16, "186f14cc94722704983db710c552f591"),  # noqa: E501
    ),
    "RemixIcon": IconFontSpec(
        IconTtfFileSpec(QtAwesome, qta_font_dir/"remixicon-2.5.0.ttf", "888e61f04316f10bddfff7bee10c6dd0"),  # noqa: E501
        IconCharMapFileSpec(QtAwesome, qta_font_dir/"remixicon-charmap-2.5.0.json", 16, "a85177747108ccaa4095606913dd6584")  # noqa: E501
    ),
}
